#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2017, California Institute of Technology
# All rights reserved.

"""
   Generate a moltemplate (.lt) file containing a definition of a polymer
   molecule whose monomers are located at the positions specified in
   "coords.raw" (a 3-column text file).  Monomers will be rotated so
   that they point in the direction connecting neighbors (r[i+1]-r[i])
   The user can specify the subunits to use when building the polymer,
   the atoms to to build bonds (and angles, and dihedrals) between monomers
   and the helical pitch of the polymer.  The output of this program is
   a text file in moltemplate (.lt) format containing the sequence of
   moltemplate commands needed to build this polymer molecule(s).  (One must
   then run moltemplate on this file to build the LAMMPS simulation files.)
       Multiple Polymers:
   To make it easier to create polymer melts, multiple polymers can be created
   from coordinates in the same file by using the "-cuts" command line argument.
      Encapsulation:
   If the "-polymer-name PolyName" command line option is given, then these
   moltemplate commands will be nested within the definition of a moltemplate
   object (named "PolyName", in this example. Later in your moltemplate files,
   you must remember to instantiate a copy of this moltemplate object using
   a command like "polymer = new PolyName"  Atoms within this object will
   share the same molecule-ID number.)  If multiple polymers are requested, then
   each of them will have their own polymer object.

"""


g_usage_msg = """
Usage:

   genpoly_lt.py  \\
      [-polymer-name pname] \\
      [-monomer-name mname] \\
      [-sequence sequence.txt] \\
      [-bond btype a1 a2] \\
      [-angle    atype a1 a2 a3 i1 i2 i3] \\
      [-dihedral dtype a1 a2 a3 a4 i1 i2 i3 i4] \\
      [-improper itype a1 a2 a3 a4 i1 i2 i3 i4] \\
      [-inherits ForceFieldObject] \\
      [-header "import \"monomer.lt\""] \\
      [-helix deltaphi] \\
      [-axis x,y,z] \\
      [-circular yes/no/connected] \\
      [-cuts cuts.txt] \\
      [-polymer-directions polarities.txt \\
      [-dir-indices ia ib] \\
      [-padding paddingX,paddingY,paddingZ] \\
      < coords.raw > polymer.lt

"""


import sys
import random
from math import *


class InputError(Exception):
    """ A generic exception object containing a string for error reporting.
        (Raising this exception implies that the caller has provided
         a faulty input file or argument.)

    """

    def __init__(self, err_msg):
        self.err_msg = err_msg

    def __str__(self):
        return self.err_msg

    def __repr__(self):
        return str(self)



class GPSettings(object):

    def __init__(self):
        self.infile_name = ''
        self.direction_orig = [1.0, 0.0, 0.0]
        self.is_circular = False
        self.connect_ends = False
        self.delta_phi = 0.0
        #self.header = 'import \"forcefield.lt\"'
        self.header = ''
        self.name_monomer = 'Monomer'
        self.name_sequence_file = ''
        self.name_polymer = ''
        self.inherits = ''
        self.dir_index_offsets = (-1,1)
        self.cuts = []
        self.reverse_polymer_directions = []
        self.box_padding = None
        self.bonds_name = []
        self.bonds_type = []
        self.bonds_atoms = []
        self.bonds_index_offsets = []
        self.angles_name = []
        self.angles_type = []
        self.angles_atoms = []
        self.angles_index_offsets = []
        self.dihedrals_name = []
        self.dihedrals_type = []
        self.dihedrals_atoms = []
        self.dihedrals_index_offsets = []
        self.impropers_name = []
        self.impropers_type = []
        self.impropers_atoms = []
        self.impropers_index_offsets = []
        self.helix_angles_file = ''
        self.orientations_file = ''
        self.orientations_use_quats = False

    def ParseArgs(self, argv):
        i = 1
        while i < len(argv):
            #sys.stderr.write('argv['+str(i)+'] = \"'+argv[i]+'\"\n')

            if ((argv[i].lower() == '-in') or
                (argv[i].lower() == '-i')):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a file name.\n')
                self.infile_name = argv[i + 1]
                del(argv[i:i + 2])
            elif argv[i].lower() == '-bond':
                if i + 3 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by 3 strings.\n')
                # self.bonds_name.append(argv[i+1])
                self.bonds_type.append(argv[i + 1])
                self.bonds_atoms.append((argv[i + 2],
                                         argv[i + 3]))
                self.bonds_index_offsets.append((0, 1))
                del(argv[i:i + 4])
            elif argv[i].lower() == '-angle':
                if i + 7 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by 4 strings and 3 integers.\n')
                # self.angles_name.append(argv[i+1])
                self.angles_type.append(argv[i + 1])
                self.angles_atoms.append((argv[i + 2],
                                          argv[i + 3],
                                          argv[i + 4]))
                self.angles_index_offsets.append((int(argv[i + 5]),
                                                  int(argv[i + 6]),
                                                  int(argv[i + 7])))
                if ((self.angles_index_offsets[-1][0] < 0) or
                    (self.angles_index_offsets[-1][1] < 0) or
                    (self.angles_index_offsets[-1][2] < 0)):
                    raise InputError(
                        'Error: ' + argv[i] + ' indices (i1 i2 i3) must be >= 0\n')
                del(argv[i:i + 8])
            elif argv[i].lower() == '-dihedral':
                if i + 9 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by 5 strings and 4 integers.\n')
                # self.dihedrals_name.append(argv[i+1])
                self.dihedrals_type.append(argv[i + 1])
                self.dihedrals_atoms.append((argv[i + 2],
                                             argv[i + 3],
                                             argv[i + 4],
                                             argv[i + 5]))
                self.dihedrals_index_offsets.append((int(argv[i + 6]),
                                                     int(argv[i + 7]),
                                                     int(argv[i + 8]),
                                                     int(argv[i + 9])))
                if ((self.dihedrals_index_offsets[-1][0] < 0) or
                    (self.dihedrals_index_offsets[-1][1] < 0) or
                    (self.dihedrals_index_offsets[-1][2] < 0) or
                    (self.dihedrals_index_offsets[-1][3] < 0)):
                    raise InputError(
                        'Error: ' + argv[i] + ' indices (i1 i2 i3 i4) must be >= 0\n')
                del(argv[i:i + 10])
            elif argv[i].lower() == '-improper':
                if i + 9 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by 5 strings and 4 integers.\n')
                # self.impropers_name.append(argv[i+1])
                self.impropers_type.append(argv[i + 1])
                self.impropers_atoms.append((argv[i + 2],
                                             argv[i + 3],
                                             argv[i + 4],
                                             argv[i + 5]))
                self.impropers_index_offsets.append((int(argv[i + 6]),
                                                     int(argv[i + 7]),
                                                     int(argv[i + 8]),
                                                     int(argv[i + 9])))
                if ((self.impropers_index_offsets[-1][0] < 0) or
                    (self.impropers_index_offsets[-1][1] < 0) or
                    (self.impropers_index_offsets[-1][2] < 0) or
                    (self.impropers_index_offsets[-1][3] < 0)):
                    raise InputError(
                        'Error: ' + argv[i] + ' indices (i1 i2 i3 i4) must be >= 0\n')
                del(argv[i:i + 10])
            elif (argv[i].lower() == '-monomer-name'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a string\n')
                self.name_monomer = argv[i + 1]
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-sequence'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a file name\n')
                self.name_sequence_file = argv[i + 1]
                del(argv[i:i + 2])

            elif (argv[i].lower() == '-cuts'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a file name\n')
                try:
                    f = open(argv[i + 1], "r")
                except IOError:
                    raise InputError(
                        'Error: File "' + argv[i + 1] + '" could not be opened for reading\n')
                for line_orig in f:
                    line = line_orig.strip()
                    ic = line.find('#')
                    if ic != -1:
                        line = line[:ic]
                    else:
                        line = line.strip()
                    if len(line) > 0:
                        try:
                            self.cuts.append(int(line))
                        except ValueError:
                            raise InputError(
                                'Error: file ' + argv[i + 1] + ' should contain only nonnegative integers.\n')
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-polymer-directions'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a file name\n')
                try:
                    f = open(argv[i + 1], "r")
                except IOError:
                    raise InputError(
                        'Error: File "' + argv[i + 1] + '" could not be opened for reading\n')
                for line_orig in f:
                    line = line_orig.strip()
                    ic = line.find('#')
                    if ic != -1:
                        line = line[:ic]
                    else:
                        line = line.strip()
                    if len(line) > 0:
                        try:
                            entry = line
                            if ((entry == '1') or
                                (entry == '+1') or
                                (entry == 'true') or
                                (entry == 'increasing')):
                                self.reverse_polymer_directions.append(False)
                            else:
                                self.reverse_polymer_directions.append(True)
                        except ValueError:
                            raise InputError(
                                'Error: file ' + argv[i + 1] + ' should contain only \"+1\" or \"-1\" on each line.\n')
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-polymer-name'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a string\n')
                self.name_polymer = argv[i + 1]
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-inherits'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a string\n')
                self.inherits = argv[i + 1]
                if self.inherits.find('inherits ') == 0:
                    self.inherits = ' ' + self.inherits
                else:
                    self.inherits = ' inherits ' + self.inherits
                if self.name_polymer == '':
                    self.name_polymer = 'Polymer'  # supply a default name
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-header'):
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a string (usually in quotes)\n')
                if self.header != '':
                    self.header += '\n'
                self.header += argv[i + 1]
                del(argv[i:i + 2])
            elif argv[i].lower() == '-axis':
                if i + 1 >= len(argv):
                    raise InputError('Error: The ' + argv[i] + ' flag should be followed ' +
                                     'by 3 numbers separated by commas (no spaces)\n')
                self.direction_orig = list(map(float, argv[i + 1].split(',')))
                del(argv[i:i + 2])
            elif argv[i].lower() == '-circular':
                if i + 1 >= len(argv):
                    raise InputError('Error: The ' + argv[i] + ' flag should be followed by an argument\n' +
                                     '       ("yes", "no", or "connected")\n')
                if argv[i + 1].lower() == 'yes':
                    self.connect_ends = True
                    self.is_circular = True
                elif argv[i + 1].lower() == 'connected':
                    self.connect_ends = True
                    self.is_circular = False
                elif argv[i + 1].lower() == 'no':
                    self.connect_ends = False
                    self.is_circular = False
                else:
                    raise InputError('Error: The ' + argv[i] + ' flag should be followed by an argument\n' +
                                     '       ("yes", "no", or "connected")\n')
                del(argv[i:i + 2])
            elif argv[i].lower() == '-helix':
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by a number (angle in degrees)\n')
                self.delta_phi = float(argv[i + 1])
                del(argv[i:i + 2])
            elif (argv[i].lower() == '-dir-indices'):
                if i + 2 >= len(argv):
                    raise InputError(
                        'Error: The ' + argv[i] + ' flag should be followed by two integers\n')
                self.dir_index_offsets = (int(argv[i + 1]), int(argv[i + 2]))
                if self.dir_index_offsets[0] == self.dir_index_offsets[1]:
                    raise InputError(
                        'Error: The two numbers following ' + argv[i] + ' must not be equal.\n')
                del(argv[i:i + 3])
            elif ((argv[i].lower() == '-padding') or
                  (argv[i].lower() == '-box')):
                if i + 1 >= len(argv):
                    raise InputError('Error: The ' + argv[i] + ' flag should be followed\n' +
                                     'by 3 numbers separated by commas (no spaces)\n')
                self.box_padding = list(map(float, argv[i + 1].split(',')))
                if len(self.box_padding) == 1:
                    self.box_padding = [self.box_padding] * 3
                del(argv[i:i + 2])
            elif (argv[i].lower() in ('-orientations','-orientations')):
                self.orientations_file = argv[i+1]
                self.orientations_use_quats = False
                del(argv[i:i + 2])
            elif (argv[i].lower() in ('-quaternion','-quaternions')):
                self.orientations_file = argv[i+1]
                self.orientations_use_quats = True
                del(argv[i:i + 2])
            elif (argv[i].lower() in ('-helix-angle','-helix-angles')):
                self.helix_angles_file = argv[i+1]
                del(argv[i:i + 2])

            # elif ((argv[i][0] == '-') and (__name__ == '__main__')):
            #
            #    raise InputError('Error('+g_program_name+'):\n'+\
            #        'Unrecogized command line argument \"'+argv[i]+\
            #        '\"\n\n'+\
            #        __doc__)
            else:
                i += 1

        if ((len(self.reverse_polymer_directions) != 0) and 
            (len(self.reverse_polymer_directions) != len(self.cuts) + 1)):
            raise InputError('Error: The number of entries in the file you supplied to "-polymer-directions"\n'
                             '       does not equal the number of polymers (which is either 1, or 1 + the\n'
                             '       number of entries in the file you supplied to "-cuts", if applicable)\n')


        for b in range(0, len(self.bonds_type)):
            if len(self.bonds_type) > 1:
                self.bonds_name.append('genp_bond' + str(b + 1) + '_')
            else:
                self.bonds_name.append('genp_bond_')
        for b in range(0, len(self.angles_type)):
            if len(self.angles_type) > 1:
                self.angles_name.append('genp_angle' + str(b + 1) + '_')
            else:
                self.angles_name.append('genp_angle_')
        for b in range(0, len(self.dihedrals_type)):
            if len(self.dihedrals_type) > 1:
                self.dihedrals_name.append('genp_dihedral' + str(b + 1) + '_')
            else:
                self.dihedrals_name.append('genp_dihedral_')
        for b in range(0, len(self.impropers_type)):
            if len(self.impropers_type) > 1:
                self.impropers_name.append('genp_improper' + str(b + 1) + '_')
            else:
                self.impropers_name.append('genp_improper_')



class WrapPeriodic(object):
    """ Wrap() calculates the remainder of i % N.
        It turns out to be convenient to do this multiple times and later
        query whether i/N != 0 in any of them once (by checking bounds_err).

    """
    bounds_err = False

    @classmethod
    def Wrap(obj, i, N):
        if i // N != 0:
            obj.bounds_err = True
        return i % N

    def WrapF(obj, x, L):
        i = floor(x / L)
        if i != 0:
            obj.bounds_err = True
        return x - i * L


class GenPoly(object):
    """
        Read coordinates from a file, and generate a list of \"new\" commands
        in moltemplate format with the position of each monomer located
        at these positions, oriented appropriately, with bonds (and angles,
        dihedrals, etc...) connecting successive monomers together.
        By default, only a single polymer is created.
        However this class can create multiple polymers of different lengths.
        The list of coordinates for each polymer are saved separately within
        the "self.coords_multi" member.

    """

    def __init__(self):
        self.settings = GPSettings()
        self.coords_multi = []  # a list-of-list-of-lists of numbers Nxnx3
        self.name_sequence_multi = []
        self.direction_vects = []
        self.box_bounds_min = None
        self.box_bounds_max = None
        self.N = 0
        self.orientations_multi = []
        self.helix_angles_multi = []


    def ParseArgs(self, argv):
        """ 
        parse the argument list
        """
        # The command above will remove arguments from argv which are
        # understood by GPSettings.ParseArgs(argv).  
        # The remaining arguments will be handled below.
        self.settings.ParseArgs(argv)



    def ReadCoords(self, infile, ncolumns=3):
        """
        Read x y z coordinates from a multi-column ASCII text file.
        Store the coordinates in the self.coords_multi[][][] list.
        (Note: Information stored in self.settings.cuts and 
               self.settings.reverse_polymer_directions 
               is used to split the coordinate list into multiple lists
               and to determine the order that coordinates are read.)
        This function can also be used to read other (multi-column) ASCII text
        files by overriding the default value for "ncolumns".  (For example,
        to read files with 4 numbers on each line, set ncolumns=4.)
        """

        filename = ''
        if isinstance(infile, str):
            filename = infile
            try:
                infile = open(filename, 'r')
            except IOError:
                raise InputError(
                    'Error: file ' + filename +
                    ' could not be opened for reading\n')

        coords_multi = []  # (do not confuse this with "self.coords_multi")
        coords = []

        lines = infile.readlines()
        for i in range(0, len(lines)):
            tokens = lines[i].strip().split()
            if (len(tokens) == ncolumns):
                coords.append(list(map(float, tokens)))

        self.N = len(coords)
        if self.N < 2:
            err_msg = 'Error: Coordinate file must have at least 2 positions'
            raise InputError(err_msg+'.\n')

        # Did the caller ask us to split the polymer into multiple polymers?
        if len(self.settings.cuts) > 0:
            if (self.settings.cuts[-1] < self.N+1):
                self.settings.cuts.append(self.N + 1)
                self.settings.cuts.sort()
            i = 0
            for j in self.settings.cuts:
                coords_multi.append(coords[i:j])
                i = j
        else:
            coords_multi.append(coords)

        # Did the caller ask us to reverse the direction of any polymers?
        for i in range(0, len(self.settings.reverse_polymer_directions)):
            assert(i < len(self.coords_multi))
            if self.settings.reverse_polymer_directions[i]:
                self.coords_multi[i].reverse()

        if filename != '':
            # Then we opened a new file with that name.  We should close it now.
            infile.close()

        return coords_multi



    def ReadSequence(self, infile):
        """
        Read a sequence of monomer type names from a file.
        This function is similar to ReadCoords().
        """
        name_sequence = []

        filename = ''
        if isinstance(infile, str):
            filename = infile
            try:
                infile = open(filename, 'r')
            except IOError:
                raise InputError(
                    'Error: file ' + filename +
                    ' could not be opened for reading\n')

        for line_orig in infile:
            line = line_orig.strip()
            ic = line.find('#')
            if ic != -1:
                line = line[:ic]
            else:
                line = line.strip()
                if len(line) > 0:
                    name_sequence.append(line)

        N = len(name_sequence)

        # Did the caller ask us to split the polymer into multiple polymers?
        if len(self.settings.cuts) > 0:
            if (self.settings.cuts[-1] < N+1):
                self.settings.cuts.append(N + 1)
                self.settings.cuts.sort()
            i = 0
            for j in self.settings.cuts:
                self.name_sequence_multi.append(name_sequence[i:j])
                i = j
        else:
            self.name_sequence_multi.append(name_sequence)

        # Did the caller ask us to reverse the direction of any polymers?
        for i in range(0, len(self.settings.reverse_polymer_directions)):
            if self.settings.reverse_polymer_directions[i]:
                self.name_sequence_multi[i].reverse()

        if filename != '':
            # Then we opened a new file with that name.  We should close it now.
            infile.close()



    def ChooseDirections(self, coords):
        """
        Calculate the direction each monomer subunit should be pointing at:

        """

        N = len(coords)

        if N == 1:
            self.direction_vects = [[1.0, 0.0, 0.0]]
            return

        self.direction_vects = [[0.0, 0.0, 0.0] for i in range(0, N + 1)]

        if self.settings.is_circular:
            for i in range(0, N):
                # By default, the direction that monomer "i" is pointing is
                # determined by the position of the monomers before and after it
                # (at index i-1, and i+1).  More generally, we allow the user
                # to choose what these offsets are ("dir_index_offsets[")
                ia = WrapPeriodic.Wrap(i + self.settings.dir_index_offsets[0],
                                       N)
                ib = WrapPeriodic.Wrap(i + self.settings.dir_index_offsets[1],
                                       N)
                for d in range(0, 3):
                    self.direction_vects[i][d] = coords[ib][d] - coords[ia][d]
                        
        else:
            for i in range(1, N - 1):
                for d in range(0, 3):
                    self.direction_vects[i][d] = (coords[i + self.settings.dir_index_offsets[1]][d]
                                                  -
                                                  coords[i + self.settings.dir_index_offsets[0]][d])

            for d in range(0, 3):
                self.direction_vects[0][d] = coords[1][d] - coords[0][d]
                self.direction_vects[N-1][d] = coords[N-1][d] - coords[N-2][d]

        # Optional: normalize the direction vectors

        for i in range(0, N):
            direction_len = 0.0
            for d in range(0, 3):
                direction_len += (self.direction_vects[i][d])**2
            direction_len = sqrt(direction_len)
            for d in range(0, 3):
                self.direction_vects[i][d] /= direction_len

        # Special case:  self.direction_vects[-1] is the direction that the original monomer
        # in "monomer.lt" was pointing.  (By default, 1,0,0 <--> the "x"
        # direction)

        self.direction_vects[-1] = self.settings.direction_orig



    def WriteLTFile(self, outfile):
        """ Write an moltemplate (.lt) file containing the definition of
        this polymer object.  (If multiple polymer objects were requested by
        the user (using the -cuts argument), then their definitions will
        appear nested within this object, and each of them will be
        instantiated once when the parent object is instantiated.)

        """

        # make sure len(genpoly.orientations_multi) == len(genpoly.coords_multi)
        if len(self.orientations_multi) == 0:
            self.orientations_multi = [[] for entry in self.coords_multi]
        # make sure len(self.helix_angles_multi) == len(self.coords_multi)
        if len(self.helix_angles_multi) == 0:
            self.helix_angles_multi = [[] for entry in self.coords_multi]

        outfile.write(self.settings.header + "\n\n\n")

        if len(self.coords_multi) == 1:
            self.WritePolymer(outfile,
                              self.settings.name_polymer +
                              self.settings.inherits,
                              self.coords_multi[0],
                              self.name_sequence_multi[0],
                              self.orientations_multi[0],
                              self.helix_angles_multi[0])
        else:
            if self.settings.name_polymer != '':
                outfile.write(self.settings.name_polymer + " {\n\n")
            outfile.write('# Definitions of individual polymers to follow\n\n')
            for i in range(0, len(self.coords_multi)):
                # insert ".../" in front of the inherits string
                ih_str = self.settings.inherits
                ih = ih_str.find('inherits ')
                if ih != -1:
                    ih += len('inherits ')
                    ih_str = ih_str[0:ih] + '.../' + ih_str[ih:]
                self.WritePolymer(outfile,
                                  self.settings.name_polymer + '_sub' + str(i + 1) +
                                  ih_str,
                                  self.coords_multi[i],
                                  self.name_sequence_multi[i],
                                  self.orientations_multi[i],
                                  self.helix_angles_multi[i])
            outfile.write('\n\n')
            outfile.write('# Now instantiate all the polymers (once each)\n\n')

            for i in range(0, len(self.coords_multi)):
                outfile.write('polymers[' + str(i) + '] = new ' +
                              self.settings.name_polymer + '_sub' + str(i + 1) + '\n')

            if self.settings.name_polymer != '':
                outfile.write('\n\n'
                              '}  # ' + self.settings.name_polymer + '\n\n')

        if self.settings.box_padding != None:
            for i in range(0, len(self.coords_multi)):
                # calculate the box big enough to collectively enclose
                # all of the coordinates (even multiple coordinate sets)
                self.CalcBoxBoundaries(self.coords_multi[i])
            self.WriteBoxBoundaries(outfile)



    def WritePolymer(self,
                     outfile,
                     name_polymer,
                     coords,
                     names_monomers,
                     orientations=[],
                     helix_angles=[]):
        """ Write a single polymer object to a file.
            This function is invoked by WriteLTFile()

        """
        N = len(coords)

        self.ChooseDirections(coords)

        if name_polymer != '':
            outfile.write(name_polymer + ' {\n'
                          '\n\n\n'
                          '  create_var {$mol}\n'
                          '  # The line above forces all monomer subunits to share the same molecule-ID\n'
                          '  # (Note: The molecule-ID number is optional and is usually ignored by LAMMPS.)\n\n\n\n')

        outfile.write("""
  # ----- list of monomers: -----
  #
  # (Note: move(), rot(), and rotvv() commands control the position
  #  of each monomer.  (See the moltemplate manual for an explanation
  #  of what they do.)  Commands enclosed in push() are cumulative
  #  and remain in effect until removed by pop().)



"""
                      )

        outfile.write("  push(move(0,0,0))\n")

        phi = 0.0

        for i in range(0, N):
            #im1 = i-1
            # if im1 < 0 or self.settings.connect_ends:
            #    if im1 < 0:
            #        im1 += N
            outfile.write("  pop()\n")
            if len(orientations) > 0:
                assert(len(orientations) == N)
                if self.settings.orientations_use_quats:
                    assert(len(orientations[i]) == 4)
                    outfile.write("  push(quat(" +
                                  str(orientations[i][0]) + "," +
                                  str(orientations[i][1]) + "," +
                                  str(orientations[i][2]) + "," +
                                  str(orientations[i][3]) + "))\n")
                else:
                    assert(len(orientations[i]) == 9)
                    outfile.write("  push(matrix(" +
                                  str(orientations[i][0]) + "," +
                                  str(orientations[i][1]) + "," +
                                  str(orientations[i][2]) + "," +
                                  str(orientations[i][3]) + "," +
                                  str(orientations[i][4]) + "," +
                                  str(orientations[i][5]) + "," +
                                  str(orientations[i][6]) + "," +
                                  str(orientations[i][7]) + "," +
                                  str(orientations[i][8]) + "))\n")
            else:
                # Otherwise, if no orientations were explicitly specified, then
                # infer the orientation from the direction of the displacement.
                outfile.write("  push(rotvv(" +
                              str(self.direction_vects[i - 1][0]) + "," +
                              str(self.direction_vects[i - 1][1]) + "," +
                              str(self.direction_vects[i - 1][2]) + "," +
                              str(self.direction_vects[i][0]) + "," +
                              str(self.direction_vects[i][1]) + "," +
                              str(self.direction_vects[i][2]) + "))\n")
            outfile.write("  push(move(" +
                          str(coords[i][0]) + "," +
                          str(coords[i][1]) + "," +
                          str(coords[i][2]) + "))\n")

            outfile.write("  mon[" + str(i) + "] = new " +
                          names_monomers[i])

            # If requested, apply additional rotations about the polymer axis
            phi = 0.0
            if len(helix_angles) > 0:
                phi += helix_angles[i]
            elif self.settings.delta_phi != 0.0:
                phi = i * self.settings.delta_phi
            # Recall that self.direction_vects[-1] =
            # self.settings.direction_orig  (usually 1,0,0)
            outfile.write(".rot(" + str(phi) +
                          "," + str(self.settings.direction_orig[0]) +
                          "," + str(self.settings.direction_orig[1]) +
                          "," + str(self.settings.direction_orig[2]) +
                          ")\n")
            if len(orientations) > 0:
                outfile.write("  pop()\n")


        outfile.write("  pop()\n")
        if len(orientations) == 0:
            outfile.write("  pop()\n"*N)

        assert(len(self.settings.bonds_name) ==
               len(self.settings.bonds_type) ==
               len(self.settings.bonds_atoms) ==
               len(self.settings.bonds_index_offsets))

        if len(self.settings.bonds_type) > 0:
            outfile.write("\n"
                          "\n"
                          "  write(\"Data Bonds\") {\n")

        WrapPeriodic.bounds_err = False
        for i in range(0, N):
            test = False
            for b in range(0, len(self.settings.bonds_type)):
                I = i + self.settings.bonds_index_offsets[b][0]
                J = i + self.settings.bonds_index_offsets[b][1]
                I = WrapPeriodic.Wrap(I, N)
                J = WrapPeriodic.Wrap(J, N)
                if WrapPeriodic.bounds_err:
                    WrapPeriodic.bounds_err = False
                    if not self.settings.connect_ends:
                        continue
                outfile.write(
                    "    $bond:" + self.settings.bonds_name[b] + str(i + 1))
                outfile.write(" @bond:" + self.settings.bonds_type[b] + " $atom:mon[" + str(I) + "]/" + self.settings.bonds_atoms[
                              b][0] + " $atom:mon[" + str(J) + "]/" + self.settings.bonds_atoms[b][1] + "\n")
        if len(self.settings.bonds_type) > 0:
            outfile.write("  }  # write(\"Data Bonds\") {...\n\n\n")


        assert(len(self.settings.angles_name) ==
               len(self.settings.angles_type) ==
               len(self.settings.angles_atoms) ==
               len(self.settings.angles_index_offsets))
        if len(self.settings.angles_type) > 0:
            outfile.write("\n"
                          "\n"
                          "  write(\"Data Angles\") {\n")
        for i in range(0, N):
            for b in range(0, len(self.settings.angles_type)):
                I = i + self.settings.angles_index_offsets[b][0]
                J = i + self.settings.angles_index_offsets[b][1]
                K = i + self.settings.angles_index_offsets[b][2]
                I = WrapPeriodic.Wrap(I, N)
                J = WrapPeriodic.Wrap(J, N)
                K = WrapPeriodic.Wrap(K, N)
                if WrapPeriodic.bounds_err:
                    WrapPeriodic.bounds_err = False
                    if not self.settings.connect_ends:
                        continue
                outfile.write(
                    "    $angle:" + self.settings.angles_name[b] + str(i + 1))
                outfile.write(" @angle:" + self.settings.angles_type[b] +
                              " $atom:mon[" + str(I) + "]/" + self.settings.angles_atoms[b][0] +
                              " $atom:mon[" + str(J) + "]/" + self.settings.angles_atoms[b][1] +
                              " $atom:mon[" + str(K) + "]/" + self.settings.angles_atoms[b][2] +
                              "\n")
        if len(self.settings.angles_type) > 0:
            outfile.write("  }  # write(\"Data Angles\") {...\n\n\n")


        assert(len(self.settings.dihedrals_name) ==
               len(self.settings.dihedrals_type) ==
               len(self.settings.dihedrals_atoms) ==
               len(self.settings.dihedrals_index_offsets))
        if len(self.settings.dihedrals_type) > 0:
            outfile.write("\n"
                          "\n"
                          "  write(\"Data Dihedrals\") {\n")
        for i in range(0, N):
            for b in range(0, len(self.settings.dihedrals_type)):
                I = i + self.settings.dihedrals_index_offsets[b][0]
                J = i + self.settings.dihedrals_index_offsets[b][1]
                K = i + self.settings.dihedrals_index_offsets[b][2]
                L = i + self.settings.dihedrals_index_offsets[b][3]
                I = WrapPeriodic.Wrap(I, N)
                J = WrapPeriodic.Wrap(J, N)
                K = WrapPeriodic.Wrap(K, N)
                L = WrapPeriodic.Wrap(L, N)
                if WrapPeriodic.bounds_err:
                    WrapPeriodic.bounds_err = False
                    if not self.settings.connect_ends:
                        continue
                outfile.write("    $dihedral:" +
                              self.settings.dihedrals_name[b] + str(i + 1))
                outfile.write(" @dihedral:" + self.settings.dihedrals_type[b] +
                              " $atom:mon[" + str(I) + "]/" + self.settings.dihedrals_atoms[b][0] +
                              " $atom:mon[" + str(J) + "]/" + self.settings.dihedrals_atoms[b][1] +
                              " $atom:mon[" + str(K) + "]/" + self.settings.dihedrals_atoms[b][2] +
                              " $atom:mon[" + str(L) + "]/" + self.settings.dihedrals_atoms[b][3] +
                              "\n")
        if len(self.settings.dihedrals_type) > 0:
            outfile.write("  }  # write(\"Data Dihedrals\") {...\n\n\n")


        assert(len(self.settings.impropers_name) ==
               len(self.settings.impropers_type) ==
               len(self.settings.impropers_atoms) ==
               len(self.settings.impropers_index_offsets))
        if len(self.settings.impropers_type) > 0:
            outfile.write("\n"
                          "\n"
                          "  write(\"Data Impropers\") {\n")
        for i in range(0, N):
            for b in range(0, len(self.settings.impropers_type)):
                I = i + self.settings.impropers_index_offsets[b][0]
                J = i + self.settings.impropers_index_offsets[b][1]
                K = i + self.settings.impropers_index_offsets[b][2]
                L = i + self.settings.impropers_index_offsets[b][3]
                I = WrapPeriodic.Wrap(I, N)
                J = WrapPeriodic.Wrap(J, N)
                K = WrapPeriodic.Wrap(K, N)
                L = WrapPeriodic.Wrap(L, N)
                if WrapPeriodic.bounds_err:
                    WrapPeriodic.bounds_err = False
                    if not self.settings.connect_ends:
                        continue
                outfile.write("    $improper:" +
                              self.settings.impropers_name[b] + str(i + 1))
                outfile.write(" @improper:" + self.settings.impropers_type[b] +
                              " $atom:mon[" + str(I) + "]/" + self.settings.impropers_atoms[b][0] +
                              " $atom:mon[" + str(J) + "]/" + self.settings.impropers_atoms[b][1] +
                              " $atom:mon[" + str(K) + "]/" + self.settings.impropers_atoms[b][2] +
                              " $atom:mon[" + str(L) + "]/" + self.settings.impropers_atoms[b][3] +
                              "\n")
        if len(self.settings.impropers_type) > 0:
            outfile.write("  }  # write(\"Data Impropers\") {...\n\n\n")

        if name_polymer != '':
            outfile.write("}  # " + name_polymer + "\n\n\n\n")

    def CalcBoxBoundaries(self, coords):
        N = len(coords)
        for i in range(0, N):
            for d in range(0, 3):
                if not self.box_bounds_min:
                    assert(not self.box_bounds_max)
                    self.box_bounds_min = [xd for xd in coords[i]]
                    self.box_bounds_max = [xd for xd in coords[i]]
                else:
                    if coords[i][d] > self.box_bounds_max[d]:
                        self.box_bounds_max[d] = coords[i][d]
                    if coords[i][d] < self.box_bounds_min[d]:
                        self.box_bounds_min[d] = coords[i][d]

    def WriteBoxBoundaries(self, outfile):
        for d in range(0, 3):
            self.box_bounds_min[d] -= self.settings.box_padding[d]
            self.box_bounds_max[d] += self.settings.box_padding[d]
        outfile.write("\n# ---------------- simulation box -----------------\n"
            
                      "# Now define a box big enough to hold a polymer with this (initial) shape\n"
                      "# (The user can override this later on.  This is the default box size.)"
                      "\n\n"
                      "write_once(\"Data Boundary\") {\n"
                      + str(self.box_bounds_min[0]) + "  " +
                      str(self.box_bounds_max[0]) + " xlo xhi\n"
                      + str(self.box_bounds_min[1]) + "  " +
                      str(self.box_bounds_max[1]) + " ylo yhi\n"
                      + str(self.box_bounds_min[2]) + "  " +
                      str(self.box_bounds_max[2]) + " zlo zhi\n"
                      "}\n\n\n")


def main():
    try:
        g_program_name = __file__.split('/')[-1]
        g_version_str = '0.1.2'
        g_date_str = '2020-8-06'
        sys.stderr.write(g_program_name + ' v' +
                         g_version_str + ' ' + g_date_str + '\n')
        argv = [arg for arg in sys.argv]
        genpoly = GenPoly()
        genpoly.ParseArgs(argv)
        # Any remain arguments?
        if len(argv) > 1:
            raise InputError('Error(' + g_program_name + '):\n' +
                             'Unrecogized command line argument \"' + argv[1] +
                             '\"\n\n' +
                             g_usage_msg)

        if genpoly.settings.infile_name != '':
            infile = open(genpoly.settings.infile_name, 'r')
        else:
            infile = sys.stdin
        outfile = sys.stdout

        # Read the coordinates
        genpoly.coords_multi = genpoly.ReadCoords(infile)

        # Did the user ask us to read a custom sequence of monomer type names?
        name_sequence_file = None
        if isinstance(genpoly.settings.name_sequence_file, str):
            if genpoly.settings.name_sequence_file != '':
                name_sequence_file = open(genpoly.settings.name_sequence_file,'r')
        if name_sequence_file:
            # Note: This will fill the contents of genpoly.name_sequence_multi
            genpoly.ReadSequence(name_sequence_file)
        else:
            # Otherwise just fill genpoly.name_sequence_multi with
            #  repeated copies of genpoly.settings.name_monomer
            #   (...using this ugly two-dimensional list-of-lists comprehension)
            genpoly.name_sequence_multi = [[genpoly.settings.name_monomer
                                            for j in
                                            range(0, len(genpoly.coords_multi[i]))]
                                           for i in range(0, len(genpoly.coords_multi))]

        # Did the user specify a file with a list of orientations?
        orientations_file = None
        if isinstance(genpoly.settings.orientations_file, str):
            if genpoly.settings.orientations_file != '':
                orientations_file = open(genpoly.settings.orientations_file, 'r')
        else:
            orientations_file = genpoly.settings.orientations_file
        if orientations_file:
            if genpoly.settings.orientations_use_quats:
                genpoly.orientations_multi = genpoly.ReadCoords(orientations_file, 4)
            else:
                genpoly.orientations_multi = genpoly.ReadCoords(orientations_file, 9)

        # Did the user specify a file with a list of helix angles?
        helix_angles_file = None
        if isinstance(genpoly.settings.helix_angles_file, str):
            if genpoly.settings.helix_angles_file != '':
                helix_angles_file = open(genpoly.settings.helix_angles_file, 'r')
        helix_angles_file = genpoly.settings.helix_angles_file
        if helix_angles_file:
            genpoly.helix_angles_multi = genpoly.ReadCoords(genpoly.settings.helix_angles_file, 1)
            # Because I borrowed the ReadCoords() function to read the file,
            # each "angle" is ended up being a list containing only one element
            # (the angle).  Convert this list into the number stored there.
            for i in range(0, len(genpoly.helix_angles_multi)):
                for j in range(0, len(genpoly.helix_angles_multi[i])):
                    assert(len(genpoly.helix_angles_multi[i][j]) == 1)
                    genpoly.helix_angles_multi[i][j] = genpoly.helix_angles_multi[i][j][0]

        # Now, check for polymer and sequence length inconsistencies:
        if (len(genpoly.coords_multi) != len(genpoly.name_sequence_multi)):
            raise InputError('Error(' +
                             g_program_name + '):\n' +
                             '      The coordinate file and sequence file have different lengths.\n')
        if ((len(genpoly.orientations_multi) > 0) and
            (len(genpoly.orientations_multi) != len(genpoly.coords_multi))):
            raise InputError('Error(' +
                             g_program_name + '):\n' +
                             '      The coordinate file and orientations/quats file have different lengths.\n')
        if ((len(genpoly.helix_angles_multi) > 0) and
            (len(genpoly.helix_angles_multi) != len(genpoly.coords_multi))):
            raise InputError('Error(' +
                             g_program_name + '):\n' +
                             '      The coordinate file and helix_angles file have different lengths.\n')
        for i in range(0, len(genpoly.coords_multi)):
            if len(genpoly.name_sequence_multi[i]) > 0:
                if (len(genpoly.coords_multi[i]) !=
                    len(genpoly.name_sequence_multi[i])):
                    raise InputError('Error(' +
                                     g_program_name + '):\n' +
                                     '      The coordinate file and sequence file have different lengths.\n')
            if ((len(genpoly.orientations_multi) > 0) and
                (len(genpoly.orientations_multi[i]) > 0)):
                if (len(genpoly.coords_multi[i]) !=
                    len(genpoly.orientations_multi[i])):
                    raise InputError('Error(' +
                                     g_program_name + '):\n' +
                                     '      The coordinate file and orientations/quats file have different lengths.\n')
            if ((len(genpoly.helix_angles_multi) > 0) and
                (len(genpoly.helix_angles_multi[i]) > 0)):
                if (len(genpoly.coords_multi[i]) !=
                    len(genpoly.helix_angles_multi[i])):
                    raise InputError('Error(' +
                                     g_program_name + '):\n' +
                                     '      The coordinate file and helix_angles file have different lengths.\n')



        # Convert all of this information to moltemplate (LT) format:
        genpoly.WriteLTFile(outfile)

        # Now close the input file
        if genpoly.settings.infile_name != '':  # <--if not reading from stdin
            infile.close()

    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
   Get rid of lines containing duplicate copies of the same atom in the "Atoms"
   section of a LAMMPS data file.  Duplicate lines which occur later are
   preserved and the earlier lines are erased.
   The file is read from sys.stdin.  This program does not parse the entire
   data file.  The text from the "Atoms" section of the LAMMPS file must
   be extracted in advance before it is sent to this program.)

"""

import sys

def main():
    in_stream = sys.stdin
    f = None
    fname = None
    if len(sys.argv) == 2:
        fname = sys.argv[1]
        f = open(fname, 'r')
        in_stream = f

    atom_ids_in_use = set([])

    lines = in_stream.readlines()

    # Start at the end of the file and read backwards.
    # If duplicate lines exist, eliminate the ones that occur earlier in the file.
    i = len(lines)
    while i > 0:
        i -= 1
        line_orig = lines[i]
        line = line_orig.rstrip('\n')
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]

        tokens = line.strip().split()
        if len(tokens) > 0:
            atom_id = tokens[0]
            if atom_id in atom_ids_in_use:
                del lines[i]
            else:
                atom_ids_in_use.add(atom_id)
        else:
            del lines[i]


    for line in lines:
        sys.stdout.write(line)

    if f != None:
        f.close()

    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Andrew Jewett (jewett.aij at g mail)
#         http://www.moltemplate.org
#         http://www.chem.ucsb.edu/~sheagroup
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013, Regents of the University of California
# All rights reserved.

try:
    from .ttree_lex import InputError
except (ImportError, SystemError, ValueError):
    # not installed as a package
    from ttree_lex import InputError

# Users of lttree typically generate the following files:
# The variable below refer to file names generated by
# write() and write_once() commands in a lttree-file.
# (I keep changing my mind what I want these names to be.)
data_prefix = "Data "
data_prefix_no_space = "Data"
data_atoms = "Data Atoms"
data_masses = "Data Masses"
data_velocities = "Data Velocities"
data_bonds = "Data Bonds"
data_bond_list = "Data Bond List"
data_angles = "Data Angles"
data_dihedrals = "Data Dihedrals"
data_impropers = "Data Impropers"
data_cmap = "Data CMAP"
data_bond_coeffs = "Data Bond Coeffs"
data_angle_coeffs = "Data Angle Coeffs"
data_dihedral_coeffs = "Data Dihedral Coeffs"
data_improper_coeffs = "Data Improper Coeffs"
data_pair_coeffs = "Data Pair Coeffs"
data_pairij_coeffs = "Data PairIJ Coeffs"

# interactions-by-type (not id. This is not part of the LAMMPS standard.)
data_bonds_by_type = "Data Bonds By Type"
data_angles_by_type = "Data Angles By Type"
data_dihedrals_by_type = "Data Dihedrals By Type"
data_impropers_by_type = "Data Impropers By Type"

# class2 data sections
data_bondbond_coeffs = "Data BondBond Coeffs"
data_bondangle_coeffs = "Data BondAngle Coeffs"
data_middlebondtorsion_coeffs = "Data MiddleBondTorsion Coeffs"
data_endbondtorsion_coeffs = "Data EndBondTorsion Coeffs"
data_angletorsion_coeffs = "Data AngleTorsion Coeffs"
data_angleangletorsion_coeffs = "Data AngleAngleTorsion Coeffs"
data_bondbond13_coeffs = "Data BondBond13 Coeffs"
data_angleangle_coeffs = "Data AngleAngle Coeffs"

# sections for non-point-like particles:
data_ellipsoids = "Data Ellipsoids"
data_lines = "Data Lines"
data_triangles = "Data Triangles"

# periodic boundary conditions
data_boundary = "Data Boundary"
# (for backward compatibility), an older version of this file was named:
data_pbc = "Data PBC"

data_file_header_names = set(['LAMMPS Description',
                              'Atoms', 'Masses', 'Velocities',
                              'Bonds', 'Angles', 'Dihedrals', 'Impropers',
                              'CMAP',
                              'Pair Coeffs', 'PairIJ Coeffs',
                              'Bond Coeffs', 'Angle Coeffs',
                              'Dihedral Coeffs', 'Improper Coeffs',
                              # class2 force fields:
                              'BondBond Coeffs', 'BondAngle Coeffs',
                              'MiddleBondTorsion Coeffs', 'EndBondTorsion Coeffs',
                              'AngleTorsion Coeffs', 'AngleAngleTorsion Coeffs',
                              'BondBond13 Coeffs',
                              'AngleAngle Coeffs',
                              # non-point-like particles:
                              'Ellipsoids', 'Triangles', 'Lines',
                              # specifying bonded interactions by type:
                              'Angles By Type', 'Dihedrals By Type', 'Impropers By Type'])

# The files are fragments of a LAMMPS data file (see "read_data").
# In addition, moltemplate may also generate the following files:
in_prefix = "In "
in_prefix_no_space = "In"
in_init = "In Init"
in_settings = "In Settings"
in_coords = "In Coords"
# These files represent different sections of the LAMMPS input script.


# Atom Styles in LAMMPS as of 2011-7-29
g_style_map = {'angle':     ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z'],
               'atomic':    ['atom-ID', 'atom-type', 'x', 'y', 'z'],
               'body':      ['atom-ID', 'atom-type', 'bodyflag', 'mass', 'x', 'y', 'z'],
               'bond':      ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z'],
               'charge':    ['atom-ID', 'atom-type', 'q', 'x', 'y', 'z'],
               'dipole':    ['atom-ID', 'atom-type', 'q', 'x', 'y', 'z', 'mux', 'muy', 'muz'],
               'dpd':       ['atom-ID', 'atom-type', 'theta', 'x', 'y', 'z'],
               'electron':  ['atom-ID', 'atom-type', 'q', 'spin', 'eradius', 'x', 'y', 'z'],
               'ellipsoid': ['atom-ID', 'atom-type', 'ellipsoidflag', 'density', 'x', 'y', 'z'],
               'full':      ['atom-ID', 'molecule-ID', 'atom-type', 'q', 'x', 'y', 'z'],
               'line':      ['atom-ID', 'molecule-ID', 'atom-type', 'lineflag', 'density', 'x', 'y', 'z'],
               'meso':      ['atom-ID', 'atom-type', 'rho', 'e', 'cv', 'x', 'y', 'z'],
               'molecular': ['atom-ID', 'molecule-ID', 'atom-type', 'x', 'y', 'z'],
               'peri':      ['atom-ID', 'atom-type', 'volume', 'density', 'x', 'y', 'z'],
               'smd':       ['atom-ID', 'atom-type', 'molecule-ID' 'volume', 'mass', 'kernel-radius', 'contact-radius', 'x', 'y', 'z'],
               'sphere':    ['atom-ID', 'atom-type', 'diameter', 'density', 'x', 'y', 'z'],
               'template':  ['atom-ID', 'molecule-ID', 'template-index', 'template-atom', 'atom-type', 'x', 'y', 'z'],
               'tri':       ['atom-ID', 'molecule-ID', 'atom-type', 'triangleflag', 'density', 'x', 'y', 'z'],
               'wavepacket': ['atom-ID', 'atom-type', 'charge', 'spin', 'eradius', 'etag', 'cs_re', 'cs_im', 'x', 'y', 'z'],
               'hybrid':    ['atom-ID', 'atom-type', 'x', 'y', 'z'],
               # The following styles were removed from LAMMPS as of 2012-3
               'colloid':   ['atom-ID', 'atom-type', 'x', 'y', 'z'],
               'granular':  ['atom-ID', 'atom-type', 'diameter', 'density', 'x', 'y', 'z']}


def AtomStyle2ColNames(atom_style_string):

    atom_style_string = atom_style_string.strip()
    if len(atom_style_string) == 0:
        raise InputError('Error: Invalid atom_style\n'
                         '       (The atom_style command was followed by an empty string.)\n')
    atom_style_args = atom_style_string.split()
    atom_style = atom_style_args[0]

    hybrid_args = atom_style_args[1:]
    if (atom_style not in g_style_map):
        if (len(atom_style_args) >= 2):
            # If the atom_style_string includes at least 2 words, then we
            # interpret this as a list of the individual column names
            return atom_style_args
        else:
            raise InputError(
                'Error: Unrecognized atom_style: \"' + atom_style + '\"\n')

    if (atom_style != 'hybrid'):
        return g_style_map[atom_style]
    else:
        column_names = ['atom-ID', 'atom-type', 'x', 'y', 'z']
        if (len(hybrid_args) == 0):
            raise InputError(
                'Error: atom_style hybrid must be followed by a sub_style.\n')
        for sub_style in hybrid_args:
            if (sub_style not in g_style_map):
                raise InputError(
                    'Error: Unrecognized atom_style: \"' + sub_style + '\"\n')
            for cname in g_style_map[sub_style]:
                if cname not in column_names:
                    column_names.append(cname)

        return column_names


def ColNames2AidAtypeMolid(column_names):
    # Because of the diversity of ways that these
    # numbers are referred to in the LAMMPS documentation,
    # we have to be flexible and allow the user to refer
    # to these quantities in a variety of ways.
    # Hopefully this covers everything:

    if 'atom-ID' in column_names:
        i_atomid = column_names.index('atom-ID')
    elif 'atom−ID' in column_names:  # (− is the character used in the manual)
        i_atomid = column_names.index('atom−ID')
    elif 'atomID' in column_names:
        i_atomid = column_names.index('atomID')
    elif 'atomid' in column_names:
        i_atomid = column_names.index('atomid')
    elif 'id' in column_names:
        i_atomid = column_names.index('id')
    elif 'atom' in column_names:
        i_atomid = column_names.index('atom')
    elif '$atom' in column_names:
        i_atomid = column_names.index('$atom')
    else:
        raise InputError('Error: List of column names lacks an \"atom-ID\"\n')

    if 'atom-type' in column_names:
        i_atomtype = column_names.index('atom-type')
    elif 'atom−type' in column_names:  # (− hyphen character used in manual)
        i_atomtype = column_names.index('atom−type')
    elif 'atomtype' in column_names:
        i_atomtype = column_names.index('atomtype')
    elif 'type' in column_names:
        i_atomtype = column_names.index('type')
    elif '@atom' in column_names:
        i_atomtype = column_names.index('@atom')
    else:
        raise InputError(
            'Error: List of column names lacks an \"atom-type\"\n')

    i_molid = None
    if 'molecule-ID' in column_names:
        i_molid = column_names.index('molecule-ID')
    elif 'molecule−ID' in column_names:  # (− hyphen character used in manual)
        i_molid = column_names.index('molecule−ID')
    elif 'moleculeID' in column_names:
        i_molid = column_names.index('moleculeID')
    elif 'moleculeid' in column_names:
        i_molid = column_names.index('moleculeid')
    elif 'molecule' in column_names:
        i_molid = column_names.index('molecule')
    elif 'molID' in column_names:
        i_molid = column_names.index('molID')
    elif 'molid' in column_names:
        i_molid = column_names.index('molid')
    elif 'mol' in column_names:
        i_molid = column_names.index('mol')
    elif '$mol' in column_names:
        i_molid = column_names.index('$mol')
    else:
        pass  # some atom_types do not have a valid molecule-ID

    return i_atomid, i_atomtype, i_molid


def ColNames2Coords(column_names):
    """ Which of the columns correspond to coordinates
        which must be transformed using rigid-body
        (affine: rotation + translation) transformations?
        This function outputs a list of lists of triplets of integers.

    """
    i_x = None
    i_y = None
    i_z = None
    if 'x' in column_names:
        i_x = column_names.index('x')
    if 'y' in column_names:
        i_y = column_names.index('y')
    if 'z' in column_names:
        i_z = column_names.index('z')
    if (((i_x != None) != (i_y != None)) or
            ((i_y != None) != (i_z != None)) or
            ((i_z != None) != (i_x != None))):
        raise InputError(
            'Error: custom atom_style list must define x, y, and z.\n')
    return [[i_x, i_y, i_z]]


def ColNames2Vects(column_names):
    """ Which of the columns correspond to coordinates
        which must be transformed using rotations?
        Some coordinates like dipole moments and
        ellipsoid orientations should only be rotated
        (not translated).
        This function outputs a list of lists of triplets of integers.

    """
    vects = []
    i_mux = None
    i_muy = None
    i_muz = None
    if 'mux' in column_names:
        i_mux = column_names.index('mux')
    if 'muy' in column_names:
        i_muy = column_names.index('muy')
    if 'muz' in column_names:
        i_muz = column_names.index('muz')
    if (((i_mux != None) != (i_muy != None)) or
            ((i_muy != None) != (i_muz != None)) or
            ((i_muz != None) != (i_mux != None))):
        raise InputError(
            'Error: custom atom_style list must define mux, muy, and muz or none.\n')
    if i_mux != None:
        vects.append([i_mux, i_muy, i_muz])
    return vects



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
#         http://www.chem.ucsb.edu/~sheagroup
# License: 3-clause BSD License  (See LICENSE.TXT)
# Copyright (c) 2013, Regents of the University of California
# All rights reserved.

man_page_text = """

    nbody_by_type.py reads a LAMMPS data file (or an excerpt of a LAMMPS)
    data file containing bonded many-body interactions by atom type
    (and bond type), and generates a list of additional interactions
    in LAMMPS format consistent with those type (to the standard out).

    Typical Usage:

    nbody_by_type.py X < old.data > new.data

    --or--

    nbody_by_type.py X \\
                     -atoms atoms.data \\
                     -bonds bonds.data \\
                     -nbody X.data \\
                     -nbodybytype X_by_type.data \\
                     > new_X.data

    In both cases "X" denotes the interaction type, which
    is either "Angles", "Dihedrals", or "Impropers".
    (Support for other interaction types can be added by the user. See below.)

    -------- Example 1 -------

    nbody_by_type.py X < old.data > new.data

    In this example, nbody_by_type.py reads a LAMMPS data file
    "orig.data", and extracts the relevant section ("Angles",
    "Dihedrals", or "Impropers").  It also looks a section named "X By Type",
       (eg. "Angles By type", "Impropers By type", "Impropers By type")
    which contains a list of criteria for automatically defining additional
    interactions of that type.  For example, this file might contain:

    Angle By Type

    7 1 2 1 * *
    8 2 2 * * *
    9 3 4 3 * *

    The first column is an interaction type ID.
    The next 3 columns are atom type identifiers.
    The final 2 columns are bond type identifiers.
    The * is a wildcard symbol indicating there is no preference for bond types
    in this example.  (Optionally, regular expressions can also be used to
    define a type match, by enclosing the atom or bond type in / slashes.)

        The first line tells us to that there should be a 3-body "Angle"
    interaction of type "7" whenever an atom of type 1 is bonded to an atom
    of type "2", which is bonded to another atom of type "1" again.
    The second line tells us that an angle is defined whenever three atoms
    are bonded together and the first two are of type "2".
    (Redundant angle interactions are filtered.)

        New interactions are created for every group of bonded
    atoms which match these criteria if they are bonded together
    in the relevant way for that interaction type (as determined by
    nbody_X.py), and printed to the standard output.  For example,
    suppose you are automatically generating 3-body "Angle" interactions using:

    nbody_by_type Angles < old.data > new.data

    The file "new.data" will be identical to "old.data", however the
    "Angles By Type" section will be deleted, and the following lines of
    text will be added to the "Angles" section:

    394 7 5983 5894 5895
    395 7 5984 5895 5896
    396 7 5985 5896 5897
     :  :   :    :    :
    847 9 14827 14848 14849

    The numbers in the first column are counters which assign a ID to
    every interaction of that type, and start where the original "Angles"
    data left off (New angle ID numbers do not overlap with old ID numbers).
    The text in the second column ("7", "9", ...) matches the text from the
    first column of the "Angle By Type" section of the input file.

    -------- Example 2 -------

    nbody_by_type.py X \\
                     -atoms atoms.data \\
                     -bonds bonds.data \\
                     -nbody X.data \\
                     -nbodybytype X_by_type.data \\
                     -prefix "SOMESTRING" -suffix "ANOTHERSTRING" \\
                     > new_X.data

    In particular, for Angle interactions:

    nbody_by_type.py Angles \\
                     -atoms atoms.data \\
                     -bonds bonds.data \\
                     -nbody angles.data \\
                     -nbodybytype angles_by_type.data \\
                     > new_Angles.data

    When run this way, nbody_by_type.py behaves exactly the same way
    as in Example 1, however only the lines of text corresponding to
    the new generated interactions are printed, (not the entire data file).
    Also note, that when run this way, nbody_by_type.py does not read the
    LAMMPS data from the standard input.  Instead, it reads each section of
    the data file from a different file indicated by the arguments following
    the "-atoms", "-bonds", "-nbody", and "-nbodybytype" flags.

    "Angles" is a 3-body interaction style.  So when run this way,
    nbody_by_type.py will create a 5 (=3+2) column file (new_Angles.data).

Note: the atom, bond and other IDs/types in need not be integers.

Note: This program must be distributed with several python modules, including:
        nbody_Angles.py, nbody_Dihedrals.py, and nbody_Impropers.py.  These
      contain bond definitions for angular, dihedral, and improper interactions.
      (In case any new interaction types are ever added to LAMMPS, with only
      a few lines of python it is easy to edit to define new bonded
      interaction types by supplying new "nbody_X.py" python module.
      Refer to the modules listed above for examples.)

Note: Optional "-prefix" and "-suffix" arguments can be included to decorate
      the interaction IDs (first column).  For example, -prefix "auto_" and
      -suffix "_angle", causes "new_Angles.data" to contain the following text:

    auto_394_angle 7 5983 5894 5895
    auto_395_angle 7 5984 5895 5896
    auto_396_angle 7 5985 5896 5897
          :        :   :    :    :
    auto_847_angle 9 14827 14848 14849

"""

g_program_name = __file__.split('/')[-1]  # = 'nbody_by_type.py'
g_date_str = '2017-2-06'
g_version_str = '0.20.0'

bond_pattern_module_name = ""

import os
import sys
sys.path.append(os.getcwd())
import importlib

if sys.version < '2.6':
    raise InputError('Error: Using python ' + sys.version + '\n'
                     '       Alas, you must upgrade to a newer version of python (2.6 or later).')
elif sys.version < '2.7':
    sys.stderr.write('--------------------------------------------------------\n'
                     '----------------- WARNING: OLD PYTHON VERSION ----------\n'
                     '  This program is untested on your python version (' +
                     sys.version + ').\n'
                     '  PLEASE LET ME KNOW IF THIS PROGRAM CRASHES (and upgrade python).\n'
                     '    -Andrew   2013-10-25\n'
                     '--------------------------------------------------------\n'
                     '--------------------------------------------------------\n')
    from ordereddict import OrderedDict
else:
    from collections import OrderedDict

try:
    from .extract_lammps_data import *
    from .nbody_by_type_lib import GenInteractions_str
    from .ttree_lex import *
    from .lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid
except (ImportError, SystemError, ValueError):
    from extract_lammps_data import *
    from nbody_by_type_lib import GenInteractions_str
    from ttree_lex import *
    from lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid



def GenInteractions_lines(lines_atoms,
                          lines_bonds,
                          lines_nbody,
                          lines_nbodybytype,
                          atom_style,
                          g_bond_pattern,
                          canonical_order,  # function to sort atoms and bonds
                          prefix='',
                          suffix='',
                          report_progress=False,
                          check_undefined=False):

    column_names = AtomStyle2ColNames(atom_style)
    i_atomid, i_atomtype, i_molid = ColNames2AidAtypeMolid(column_names)

    atomids_str = []
    atomtypes_str = []

    for iv in range(0, len(lines_atoms)):
        line = lines_atoms[iv].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = SplitQuotedString(line)
            if ((len(tokens) <= i_atomid) or (len(tokens) <= i_atomtype)):
                raise(InputError('Error not enough columns on line ' +
                                 str(iv + 1) + ' of \"Atoms\" section.'))
            tokens = SplitQuotedString(line)
            atomids_str.append(EscCharStrToChar(tokens[i_atomid]))
            atomtypes_str.append(EscCharStrToChar(tokens[i_atomtype]))

    bondids_str = []
    bondtypes_str = []
    bond_pairs = []
    for ie in range(0, len(lines_bonds)):
        line = lines_bonds[ie].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = SplitQuotedString(line)
            if len(tokens) < 4:
                raise(InputError('Error not enough columns on line ' +
                                 str(ie + 1) + ' of \"Bonds\" section.'))
            bondids_str.append(EscCharStrToChar(tokens[0]))
            bondtypes_str.append(EscCharStrToChar(tokens[1]))
            bond_pairs.append((EscCharStrToChar(tokens[2]),
                               EscCharStrToChar(tokens[3])))

    typepattern_to_coefftypes = []

    for i in range(0, len(lines_nbodybytype)):
        line = lines_nbodybytype[i].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = SplitQuotedString(line)

            if ((len(tokens) != 1 + g_bond_pattern.GetNumVerts()) and
                (len(tokens) != 1 + g_bond_pattern.GetNumVerts()
                                  + g_bond_pattern.GetNumEdges())):
                raise(InputError('Error: Wrong number of columns in \"By Type\" section of data file.\n'
                                 'Offending line:\n' +
                                 '\"' + line + '\"\n'
                                 'Expected either ' +
                                 str(1 + g_bond_pattern.GetNumVerts()) + ' or '+
                                 str(1 + g_bond_pattern.GetNumVerts() +
                                     g_bond_pattern.GetNumEdges())
                                 + ' colunms.'))

            coefftype = EscCharStrToChar(tokens[0])
            typepattern = []

            for typestr in tokens[1:]:
                if ((len(typestr) >= 2) and
                    HasRE(typestr)):
                    regex_str = typestr[3:]
                    typepattern.append(re.compile(regex_str))
                else:
                    typepattern.append(EscCharStrToChar(typestr))

            # If the user neglected to specify the bond types, assume '*'
            if len(tokens) == 1 + g_bond_pattern.GetNumVerts():
                typepattern += ['*'] * g_bond_pattern.GetNumEdges()

            typepattern_to_coefftypes.append([typepattern, coefftype])

    coefftype_to_atomids_str = GenInteractions_str(bond_pairs,
                                                   g_bond_pattern,
                                                   typepattern_to_coefftypes,
                                                   canonical_order,
                                                   atomids_str,
                                                   atomtypes_str,
                                                   bondids_str,
                                                   bondtypes_str,
                                                   report_progress,
                                                   check_undefined)
    lines_nbody_new = []
    for coefftype, atomids_list in coefftype_to_atomids_str.items():
        for atomids_found in atomids_list:
            n = len(lines_nbody) + len(lines_nbody_new) + 1
            line = prefix + str(n) + suffix + ' ' + \
                coefftype + ' ' + (' '.join(atomids_found)) + '\n'
            lines_nbody_new.append(line)

    return lines_nbody_new


def GenInteractions_files(lines_data,
                          src_bond_pattern,
                          fname_atoms,
                          fname_bonds,
                          fname_nbody,
                          fname_nbodybytype,
                          section_name,
                          section_name_bytype,
                          atom_style,
                          prefix='',
                          suffix='',
                          report_progress=False,
                          check_undefined=False):

    if fname_atoms == None:
        lines_atoms = [
            line for line in ExtractDataSection(lines_data, 'Atoms')]
    else:
        try:
            f = open(fname_atoms, 'r')
        except:
            sys.stderr.write('Error: Unable to open file \"' +
                             fname_atoms + '\" for reading.\n')
            sys.exit(-1)
        lines_atoms = [line for line in f.readlines()
                       if ((len(line.strip()) > 0) and (line.strip()[0] != '#'))]
        f.close()

    if fname_bonds == None:
        lines_bonds = [
            line for line in ExtractDataSection(lines_data, 'Bonds')]
    else:
        try:
            f = open(fname_bonds, 'r')
        except IOError:
            sys.stderr.write('Error: Unable to open file \"' +
                             fname_bonds + '\" for reading.\n')
            sys.exit(-1)
        lines_bonds = [line for line in f.readlines()
                       if ((len(line.strip()) > 0) and (line.strip()[0] != '#'))]
        f.close()

    if fname_nbody == None:
        lines_nbody = [line for line in ExtractDataSection(
            lines_data, section_name)]
    else:
        try:
            f = open(fname_nbody, 'r')
            lines_nbody = [line for line in f.readlines()
                           if ((len(line.strip()) > 0) and (line.strip()[0] != '#'))]
            f.close()
        except IOError:
            #sys.stderr.write('    (omitting optional file \"'+fname_nbody+'\")\n')
            lines_nbody = []

    if fname_nbodybytype == None:
        lines_nbodybytype = [line for
                             line in ExtractDataSection(lines_data,
                                                        section_name_bytype)]

    else:
        try:
            f = open(fname_nbodybytype, 'r')
        except:
            sys.stderr.write('Error: Unable to open file \"' +
                             fname_nbodybytype + '\" for reading.\n')
            sys.exit(-1)
        lines_nbodybytype = [line for line in f.readlines()
                             if((len(line.strip()) > 0)and(line.strip()[0] != '#'))]
        f.close()

    # search locations
    package_opts = [[src_bond_pattern, __package__],
                    ['nbody_alt_symmetry.'+src_bond_pattern, __package__]]

    if __package__:
        for i in range(0, len(package_opts)):
            package_opts[i][0] = '.' + package_opts[i][0]
        package_opts.append(['.'+src_bond_pattern, __package__+'.nbody_alt_symmetry'])


    g = None
    for name, pkg in package_opts:
        try:
            g = importlib.import_module(name, pkg)
            break
        except (ImportError, SystemError, ValueError):
            pass

    if g is None:
        sys.stderr.write('Error: Unable to locate file \"' +
                         src_bond_pattern + '.py\"\n'
                         '       (Did you mispell the file name?\n'
                         '        Check the \"nbody_alt_symmetry/\" directory.)\n')
        sys.exit(-1)

    return GenInteractions_lines(lines_atoms,
                                 lines_bonds,
                                 lines_nbody,
                                 lines_nbodybytype,
                                 atom_style,
                                 g.bond_pattern,
                                 g.canonical_order,
                                 prefix,
                                 suffix,
                                 report_progress,
                                 check_undefined)


def main():
    sys.stderr.write(g_program_name + ' v' +
                     g_version_str + ' ' + g_date_str + ' ')
    if sys.version < '3':
        sys.stderr.write(' (python version < 3)\n')
    else:
        sys.stderr.write('\n')

    try:

        fname_atoms = None
        fname_bonds = None
        fname_nbody = None
        fname_nbodybytype = None
        atom_style = 'full'
        prefix = ''
        suffix = ''
        check_undefined = False

        argv = [arg for arg in sys.argv]

        if len(argv) == 1:
            raise InputError('Error: Missing argument required.\n'
                             '       The \"' + g_program_name +
                             '\" program requires an argument containing the\n'
                             '       name of a section from a LAMMPS data file storing bonded interactions.\n'
                             '       (For example: "Angles", "Dihedrals", or "Impropers".)\n'
                             #'        Note: The first letter of each section is usually capitalized.)\n'
                             '\n'
                             '--------------- general documentation -------------\n'
                             '\n' + man_page_text + '\n')

        section_name = ''         # (This will be replaced later.)
        section_name_bytype = ''  # (This will be replaced later.)

        # Loop over the remaining arguments not processed yet.
        # These arguments are specific to the lttree.py program
        # and are not understood by ttree.py:
        i = 1
        while i < len(argv):
            #sys.stderr.write('argv['+str(i)+'] = \"'+argv[i]+'\"\n')
            if ((argv[i].lower() == '-?') or
                    (argv[i].lower() == '--?') or
                    (argv[i].lower() == '-help') or
                    (argv[i].lower() == '-help')):
                if i + 1 >= len(argv):
                    sys.stdout.write(man_page_text + '\n')
                    sys.exit(0)

            elif argv[i].lower() == '-atoms':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                     '       text which might appear in the "Atoms" section of a LAMMPS data file.\n')
                fname_atoms = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bonds':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                     '       text which might appear in the "Bonds" section of a LAMMPS data file.\n')
                fname_bonds = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-nbody':
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')

                    # raise InputError('Error: '+argv[i]+' flag should be followed by a file name containing lines of\n'
                    #                 '       text which might appear in the "'+section_name+' section of a LAMMPS data file.\n')
                fname_nbody = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-nbodybytype':
                if i + 1 >= len(argv):
                    raise InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')

                    # raise InputError('Error: '+argv[i]+' flag should be followed by a file name containing\n'
                    #                 '       text which might appear in the "'+section_name+' By Type" section\n'
                    #                 '       of a LAMMPS data file.\n')
                fname_nbodybytype = argv[i + 1]
                del(argv[i:i + 2])

            elif ((argv[i].lower() == '-atom-style') or
                  (argv[i].lower() == '-atom_style')):
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by a an atom_style name.\n'
                                     '       (Or single quoted string which includes a space-separated\n'
                                     '       list of column names.)\n')
                atom_style = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-prefix':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by a prefix string\n'
                                     '       (a string you want to appear to the left of the integer\n'
                                     '        which counts the bonded interactions you have generated.)\n')
                prefix = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-suffix':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by a suffix string\n'
                                     '       (a string you want to appear to the right of the integer\n'
                                     '        which counts the bonded interactions you have generated.)\n')
                prefix = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-subgraph':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by the name of a python file\n'
                                     '       containing the definition of the subgraph you are searching for\n'
                                     '       and it\'s symmetry properties.\n'
                                     '       (See nbody_Dihedrals.py for example.)\n')
                bond_pattern_module_name = argv[i + 1]
                # If the file name ends in ".py", then strip off this suffix.
                # The next line does not work.  Too lazy to care why.
                # bond_pattern_module_name=bond_pattern_module_name.rstrip('.py')
                # Do this instead
                pc = bond_pattern_module_name.rfind('.py')
                if pc != -1:
                    bond_pattern_module_name = bond_pattern_module_name[0:pc]

                del(argv[i:i + 2])

            elif argv[i].lower() == '-section':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by the name of the LAMMPS\n'
                                     '       Data section describing the type of interaction being generated.\n'
                                     '       (For example: \"Angles\", \"Dihedrals\", \"Impropers\", etc...)\n')
                section_name = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-sectionbytype':
                if i + 1 >= len(argv):
                    raise InputError('Error: ' + argv[i] + ' flag should be followed by the name of the\n'

                                     '       write_once(\"???? By Type\") section describing how to create the\n'
                                     '       interactions.  (For example: \"Angles By Type\", \"Dihedrals By Type\",\n'
                                     '        \"Impropers By Type\", etc...  Note that this argument\n'
                                     '        will contain spaces, so surround it with quotes.)\n')

                section_name_bytype = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-checkff':
                check_undefined = True
                del(argv[i:i + 1])

            elif argv[i][0] == '-':
                raise InputError('Error(' + g_program_name + '):\n'
                                 'Unrecogized command line argument \"' + argv[i] + '\"\n')
            else:
                i += 1

        # if len(argv) == 1:
        #    raise InputError('Error: Missing argument required.\n'
        #                     '       The \"'+g_program_name+'\" program requires an argument containing the\n'
        #                     '       name of a section from a LAMMPS data file storing bonded interactions.\n'
        #                     '       (For example: "Angles", "Dihedrals", or "Impropers".)\n')
        #                     #'        Note: The first letter of each section is usually capitalized.)\n'

        if len(argv) == 1:
            pass
        elif len(argv) == 2:
            section_name = argv[1]
            section_name_bytype = section_name + ' By Type'
            # default bond_pattern_module name
            if bond_pattern_module_name == "":  # <--if not set by user
                bond_pattern_module_name = 'nbody_' + section_name
            del(argv[1:2])
        else:
            # if there are more than 2 remaining arguments,
            problem_args = ['\"' + arg + '\"' for arg in argv[1:]]
            raise InputError('Syntax Error(' + g_program_name + '):\n\n'
                             '       Problem with argument list.\n'
                             '       The remaining arguments are:\n\n'
                             '         ' + (' '.join(problem_args)) + '\n\n'
                             '       (The actual problem may be earlier in the argument list.)\n')

        if ((section_name == '') or
            (section_name_bytype == '') or
            (bond_pattern_module_name == '')):
            raise InputError('Syntax Error(' + g_program_name + '):\n\n'
                             '       You have not defined the following arguments:\n'
                             '       -section name\n'
                             '       -sectionbytype namebytype\n'
                             '       -subgraph pythonfile.py\n')

        # ------------ Done parsing argument list ----------

        if (fname_atoms or fname_bonds or fname_nbody or fname_nbodybytype):
            output_full_DATA_file = False
            lines_data = []
        else:
            output_full_DATA_file = True
            lines_data = sys.stdin.readlines()

        # Calculate the interactions and generate a list of lines of text

        lines_new_interactions = \
            GenInteractions_files(lines_data,
                                  bond_pattern_module_name,
                                  fname_atoms,
                                  fname_bonds,
                                  fname_nbody,
                                  fname_nbodybytype,
                                  section_name,
                                  section_name_bytype,
                                  atom_style,
                                  prefix,
                                  suffix,
                                  True,
                                  check_undefined)

        # Print this text to the standard out.

        # Question: Do we write out the entire DATA file,
        # or just the portion that was generated by this program?

        if not output_full_DATA_file:
            # ...then only print out the interactions which were generated
            # by this program, omitting any lines from the original data file:

            # (This is the way I usually run this program.)
            for line in lines_new_interactions:
                sys.stdout.write(line)

        else:

            # ...then print out the entire data file, deleting the "By Type"
            # section, and adding the generated lines of text to the
            # corresponding

            # If present, update the interaction counter at the beginning
            # of the LAMMPS data file.  (For example, if if 100 new "Angles"
            # interactions were generated, replace "2 Angles" with "102 Angles")
            #
            for i in range(0, len(lines_data)):
                line = lines_data[i].strip()
                tokens = SplitQuotedString(line)

                # updating the interaction counter
                if ((len(tokens) == 2) and (tokens[1] == (section_name).lower())):
                    tokens[0] = str(int(tokens[0]) +
                                    len(lines_new_interactions))
                    lines_data[i] = ' '.join(tokens) + '\n'

                # stop when you come to a section header
                elif line in lammps_data_sections:
                    #"lammps_data_sections" is defined in "extract_lammps_data.py"
                    break

            # locate the appropriate section of the data file
            # (storing the type of interactions we just created)
            i_nbody_a, i_nbody_b = \
                FindDataSection(lines_data, section_name)

            if i_nbody_a == -1:
                if len(lines_new_interactions) > 0:
                    # If not found, create a new section at the end of the file,
                    # containing a section name followed by the list of lines
                    lines_data += ['\n', section_name + '\n', '\n'] + \
                        lines_new_interactions + ['\n']
            else:
                # Insert the new lines into the existing section
                lines_data[i_nbody_b:i_nbody_b] = lines_new_interactions

            # Figure out where the "By Type" section is located
            # (so we skip over it)
            i_bytype_a, i_bytype_b = \
                FindDataSection(lines_data, section_name_bytype)

            in_bytype_section = False
            for i in range(0, len(lines_data)):
                line = lines_data[i].strip()
                # Omit all lines of text in the 'By Type' section (including the
                # header and commments or blank lines which immediately follow
                # it.)
                if line == section_name_bytype:
                    in_bytype_section = True
                elif i == i_bytype_b:
                    in_bytype_section = False

                if not in_bytype_section:
                    sys.stdout.write(lines_data[i])

    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2017, California Institute of Technology
# All rights reserved.

g_program_name = __file__.split('/')[-1]  # = 'interpolate_curve.py'
g_version_str = '0.3.1'
g_date_str = '2020-4-11'

g_usage_str = """
Usage:

   """ + g_program_name + """ Ndesired [scale] [alpha] < coords_orig.raw > coords.raw

Example:

   """ + g_program_name + """ 30117 3.0 0.5 < coords_orig.raw > coords.raw

"""


import sys
from math import *
import numpy as np
import bisect



## Tri Diagonal Matrix Algorithm(a.k.a Thomas algorithm) solver stolen from:
# https://gist.github.com/cbellei/8ab3ab8551b8dfc8b081c518ccd9ada9
# (also taken from https://gist.github.com/ofan666/1875903)

import numpy as np

## Tri Diagonal Matrix Algorithm(a.k.a Thomas algorithm) solver
def TDMAsolver(a, b, c, d):
    '''
    TDMA solver, a b c d can be NumPy array type or Python list type.
    Solves a (non-cyclic) system of equations of the form:

      a_i*x_{i-1} + b_i*x_i + c_i*x_{i+1} = d_i
      where a_1=0, and c_n=0

    refer to http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm
    and to http://www.cfd-online.com/Wiki/Tridiagonal_matrix_algorithm_-_TDMA_(Thomas_algorithm)
    '''
    nf = len(d) # number of equations
    ac, bc, cc, dc = map(np.array, (a, b, c, d)) # copy arrays
    for it in range(1, nf):
        mc = ac[it-1]/bc[it-1]
        bc[it] = bc[it] - mc*cc[it-1] 
        dc[it] = dc[it] - mc*dc[it-1]
        	    
    xc = bc
    xc[-1] = dc[-1]/bc[-1]

    for il in range(nf-2, -1, -1):
        xc[il] = (dc[il]-cc[il]*xc[il+1])/bc[il]

    return xc



def CalcNaturalCubicSplineCoeffs(r, spline_exponent_alpha=0.5):
    N = len(r)
    assert(N >= 4)
    D = len(r[0])
    tcontrol = np.zeros(N)

    # e_d2rdt2[d][i] is the second derivative of the spline at the ith 
    #               control point (and in the dth direction)
    #
    # Once we have figured out e_d2rdt2[d][i], we can calculate the spline
    # anywhere using:
    #   SplineEval(t, h_i, 
    #              e_{i+1} / (6*h_i), 
    #              e_i / (6*h_{i+1}),
    #              r_{i+1}/h_i - e_{i+1}*h_i,
    #              r_i/h_i - e_i*h_i)

    e_d2rdt2 = np.zeros((D,N))

    # We want to solve this system of equations for e_1, e_2, ..., e_{n-2}:
    #    (note: e_i is shorthand for e_d2rdt2[d][i])
    #
    # h_{i-1}*e_{i-1} + u_i*e_i + h_{i+1} * e_{i+1}  =  v_i
    #   where h_i, u_i and v_i are shorthand for:
    #
    # h_i = change in "time" parameter for the ith interval, which is usually:
    #     = |r_{i+1} - r_i |^alpha    (alpha varies depending on settings)
    #   and
    # u_i = 2 * (h_{i-1} + h_i )
    # v_i = 6 * (b_i     - b_{i-1})
    # b_i =  (r_i - r_{i-1}) / h_i
    #
    # ...subject to the constraints that the curve is straight at the endpoints:
    # e_0 = 0       <-- first control point  (indexing begins at 0)
    # e_{n-1} = 0   <-- this is the last control point  (indexing begins at 0)

    h_dt = np.zeros(N-1)   # h_dt[i] is the i'th time interval 
                         # in the parameterization
    b_drdt = np.zeros((N-1,D))  # b_drdt is a discrete version of the derivative

    t_total=0.0
    for i in range(0, N-1):
        tcontrol[i] = t_total
        sqr_distance_i_ip1 = 0.0
        for d in range(0, D):
            sqr_distance_i_ip1 += (r[i+1][d] - r[i][d])**2
        h_dt[i] = sqr_distance_i_ip1**(0.5*spline_exponent_alpha)
        for d in range(0, D):
            b_drdt[i][d] = (r[i+1][d] - r[i][d]) / h_dt[i]
        t_total += h_dt[i]
    tcontrol[N-1] = t_total

    a_coeff = np.zeros(N)
    b_coeff = np.zeros(N)
    c_coeff = np.zeros(N)
    d_coeff = np.zeros((D,N))

    for i in range(1, N-1):
        # h_dt[i] is the difference in "time" in the parametric curve 
        # between pairs of control points.  If spline_exponenent_alpha is 0
        # then the time interval between control points is uniform.
        # ("spline_exponent_alpha" is 0.5 for centripital Catmull-Rom splines.)
        a_coeff[i]    =      h_dt[i-1]
        b_coeff[i]    = 2.0*(h_dt[i-1] + h_dt[i])
        c_coeff[i]    =      h_dt[i]
        for d in range(0, D):
            d_coeff[d][i] = 6.0*(b_drdt[i][d] - b_drdt[i-1][d])

    a_coeff[0] = 0.0
    b_coeff[0] = 1.0
    c_coeff[0] = 0.0
    for d in range(0, D):
        d_coeff[d][0] = 0.0
    a_coeff[N-1] = 0.0
    b_coeff[N-1] = 1.0
    c_coeff[N-1] = 0.0
    for d in range(0, D):
        d_coeff[d][N-1] = 0.0

    for d in range(0, D):
        e_d2rdt2[d] = TDMAsolver(a_coeff, b_coeff, c_coeff, d_coeff[d])

    # alternately, if that fails, try the matrix inverter that comes with numpy:
    #M = np.zeros((N,N))
    #for i in range(0,N):
    #    if i-1>=0:
    #        M[i][i-1] = a_coeff[i]
    #    M[i][i]   = b_coeff[i]
    #    if i+1 < N:
    #        M[i][i+1] = c_coeff[i]
    #for d in range(0, D):
    #    e_d2rdt2[d] = np.linalg.solve(M, d_coeff[d])

    e_d2rdt2 = e_d2rdt2.transpose()
    c3a = np.zeros((N, D))
    c3b = np.zeros((N, D))
    c1a = np.zeros((N, D))
    c1b = np.zeros((N, D))
    # faster to precompute these coefficients in advance:
    for i in range(0, N-1):
        # c3a = e_{i+1} / (6*h_i)
        # c3b =     e_i / (6*h_{i+1})
        # c1a = r_{i+1}/h_i - e_{i+1}*h_i
        # c1b =  r_i / h_i - e_i*h_i)
        c3a[i] = e_d2rdt2[i+1] / (6*h_dt[i])
        c3b[i] = e_d2rdt2[i]   / (6*h_dt[i])
        c1a[i] = r[i+1]/h_dt[i]  -  e_d2rdt2[i+1]*h_dt[i]/6.0
        c1b[i] = r[i]/h_dt[i]    -  e_d2rdt2[i]*h_dt[i]/6.0

    # Return these spline coefficients to the caller.
    # We can use these to quickly evaluate the spline repetatively later on

    return c3a, c3b, c1a, c1b, tcontrol





def SplineEval(t, t_interval, c3a, c3b, c1a, c1b):
    ta = t
    tb = t_interval - t
    return (c3a*ta*ta + c1a)*ta + (c3b*tb*tb + c1b)*tb



def SplineEvalD1(t, t_interval, c3a, c3b, c1a, c1b):
    ta = t
    tb = t_interval - t
    return c3a*ta*ta*3.0 + c1a - (c3b*tb*tb*3.0 + c1b)



def SplineEvalD2(t, t_interval, c3a, c3b, c1a, c1b):
    ta = t
    tb = t_interval - t
    return c3a*ta*6.0 + c3b*tb*6.0



def SplineInterpEval(t, c3a, c3b, c1a, c1b, tcontrol):
    i = bisect.bisect(tcontrol, t) - 1
    if i < 0: i = 0
    if i >= len(tcontrol)-1: i = len(tcontrol)-2
    return SplineEval(t-tcontrol[i], tcontrol[i+1]-tcontrol[i],
                      c3a[i], c3b[i], c1a[i], c1b[i])


def SplineInterpEvalD1(t, c3a, c3b, c1a, c1b, tcontrol):
    i = bisect.bisect(tcontrol, t) - 1
    if i < 0: i = 0
    if i >= len(tcontrol)-1: i = len(tcontrol)-2
    return SplineEvalD1(t-tcontrol[i], tcontrol[i+1]-tcontrol[i],
                        c3a[i], c3b[i], c1a[i], c1b[i])



def SplineInterpEvalD2(t, c3a, c3b, c1a, c1b, tcontrol):
    i = bisect.bisect(tcontrol, t) - 1
    if i < 0: i = 0
    if i >= len(tcontrol)-1: i = len(tcontrol)-2
    return SplineEvalD2(t-tcontrol[i], tcontrol[i+1]-tcontrol[i],
                        c3a[i], c3b[i], c1a[i], c1b[i])



def SplineCurvature2D(t, t_interval, c3a, c3b, c1a, c1b):
    # first derivatives
    x1,y1 = SplineEvalD1(t, t_interval, c3a, c3b, c1a, c1b)
    # second derivatives
    x2,y2 = SplineEvalD2(t, t_interval, c3a, c3b, c1a, c1b)
    denom = pow((x1*x1 + y1*y1), 1.5)
    curvature = (x1*y2 - x2*y1) / denom
    return curvature



def SplineInterpCurvature2D(t, c3a, c3b, c1a, c1b, tcontrol):
    i = bisect.bisect(tcontrol, t) - 1
    if i < 0: i = 0
    if i >= len(tcontrol)-1: i = len(tcontrol)-2
    return SplineCurvature2D(t-tcontrol[i], tcontrol[i+1]-tcontrol[i],
                             c3a[i], c3b[i], c1a[i], c1b[i])



def ResampleCurve(x_orig, num_points, alpha=0.5):
    """
       Given a list (or numpy array) of points in n-dimensional space that lie
    along some curve, this function returns a new list of "num_points" points
    which are uniformly distributed along the original curve.  This is done
    using cubic spline interpolation (which is tuned by the "alpha" parameter).

    https://en.wikipedia.org/wiki/Cubic_Hermite_spline#Catmull%E2%80%93Rom_spline

       Here "uniformly" means distributed at even intervals in the spline-
    parameter-space, not in physical distance. (If the original points were
    not uniformly distributed along the curve, then new points won't be either.)

       This function has not been optimized for speed.
    """

    assert(len(x_orig) >= 4)
    num_dimensions = len(x_orig[0])
    x_new = np.zeros((num_points, num_dimensions))

    c3a, c3b, c1a, c1b, tcontrol = \
        CalcNaturalCubicSplineCoeffs(x_orig, alpha)
    tmin = 0.0
    tmax = tcontrol[-1]
    for i in range(0, num_points):
        t = tmin + i*((tmax - tmin) / (num_points-1))
        x_new[i] = SplineInterpEval(t, c3a, c3b, c1a, c1b, tcontrol)

    return x_new


class InputError(Exception):
    """ A generic exception object containing a string for error reporting.
        (Raising this exception implies that the caller has provided
         a faulty input file or argument.)

    """

    def __init__(self, err_msg):
        self.err_msg = err_msg

    def __str__(self):
        return self.err_msg

    def __repr__(self):
        return str(self)


def main():

    try:

        #######  Main Code Below: #######
        sys.stderr.write(g_program_name+' v'+g_version_str+' '+g_date_str+'\n')

        spline_exponent_alpha = 0.5
        use_linear_interpolation = False

        # Parse the argument list:
        if len(sys.argv) <= 1:
            raise InputError("Missing arguments\n"+g_usage_str+"\n")
                             

        n_new = int(sys.argv[1])

        if len(sys.argv) > 2:
            scale = float(sys.argv[2])
        else:
            scale = 1.0

        if len(sys.argv) > 3:
            spline_exponent_alpha = float(sys.argv[3])            

        coords_orig = []

        lines = sys.stdin.readlines()

        for line in lines:
            tokens = line.split()
            if (len(tokens) > 0):
                coords_orig.append(list(map(float, tokens)))
                g_dim = len(tokens)

        n_orig = len(coords_orig)

        x_orig = np.array(coords_orig)

        if n_orig < 2:
            raise InputError("Input file contains less than two lines of coordinates.")

        if n_orig < 4:
            use_linear_interpolation = True

        if n_new < 2:
            raise InputError("Output file will contain less than two lines of coordinates.")

        #coords_new = [[0.0 for d in range(0, g_dim)] for i in range(0, n_new)]
        x_new = np.zeros((n_new, g_dim))

        if use_linear_interpolation:
            for i_new in range(0, n_new):
                I_orig = (i_new) * (float(n_orig-1) / float(n_new-1))
                i_orig = int(floor(I_orig))
                i_remainder = I_orig - i_orig

                if (i_new < n_new-1):
                    for d in range(0, g_dim):
                        x_new[i_new][d] = (x_orig[i_orig][d]
                                           +
                                           i_remainder*(x_orig[i_orig+1][d]-
                                                        x_orig[i_orig][d]))
                else:
                    for d in range(0, g_dim):
                        x_new[i_new][d] = x_orig[n_orig-1][d]

        else:
            x_new = ResampleCurve(x_orig, n_new, spline_exponent_alpha)

        # print the coordates
        for i in range(0, n_new):
            for d in range(0, g_dim-1):
                sys.stdout.write(str(x_new[i][d]*scale) + ' ')
            sys.stdout.write(str(x_new[i][g_dim-1]*scale) + "\n")

    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + 'Error:\n\n' + str(err) + '\n')
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

"""
    bonds_by_type.py reads a LAMMPS data file (or an excerpt of a LAMMPS)
    data file containing bonded many-body interactions by atom type
    (and bond type), and generates a list of additional interactions
    in LAMMPS format consistent with those type (to the standard out).

    Typical Usage:

    bonds_by_type.py -atoms atoms.data \\
                     -bonds bonds.data \\
                     -bondsbytype bonds_by_type.data \\
                     > new_bonds.data

"""


g_program_name = __file__.split('/')[-1]  # = 'bonds_by_type.py'
g_date_str = '2016-12-21'
g_version_str = '0.12.0'


import sys

try:
    from . import ttree_lex
    from .lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid
except (ImportError, SystemError, ValueError):
    # not installed as a package
    import ttree_lex
    from lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid


def LookupBondTypes(bond_types,
                    bond_ids,
                    bond_pairs,
                    lines_atoms,
                    lines_bonds,
                    lines_bondsbytype,
                    atom_style,
                    section_name,
                    prefix='',
                    suffix='',
                    bond_ids_offset=0):
                    # report_progress = False):
    """
    LookupBondTypes() looks up bond types.

    Output:
    ...It looks up the corresponding type of each bond and store it in the
       "bond_types" list.  (If the bond_ids were not specified by the user,
       generate them and store them in the bond_ids list.)


    Input (continued):
       This function requires:
    ...a list of bonded pairs of atoms
        stored in the lines_bonds variable (from the "Data Bond List"
                                       or "Data Bonds AtomId AtomId" sections)
    ...and a list of atom types
        stored in the lines_atoms variable (from the "Data Atoms" section)
    ...and a list of bond-types-as-a-function-of-atom-types
        stored in the lines_bondsbytype (from the "Data Bonds By Type" section)

    Generated bond_ids (if applicable) are of the form
      prefix + str(number) + suffix
        (where "number" begins at bond_ids_offset+1)

    """

    column_names = AtomStyle2ColNames(atom_style)
    i_atomid, i_atomtype, i_molid = ColNames2AidAtypeMolid(column_names)

    atomids = []
    atomtypes = []
    atomids2types = {}

    for iv in range(0, len(lines_atoms)):
        line = lines_atoms[iv].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = ttree_lex.SplitQuotedString(line)
            if ((len(tokens) <= i_atomid) or (len(tokens) <= i_atomtype)):
                sys.stderr.write("\"" + line + "\"\n")
                raise(ttree_lex.InputError(
                    'Error not enough columns on line ' + str(iv + 1) + ' of \"Atoms\" section.'))
            tokens = ttree_lex.SplitQuotedString(line)
            atomid = ttree_lex.EscCharStrToChar(tokens[i_atomid])
            atomids.append(atomid)
            atomtype = ttree_lex.EscCharStrToChar(tokens[i_atomtype])
            atomtypes.append(atomtype)
            atomids2types[atomid] = atomtype

    assert(isinstance(bond_ids, list))
    assert(isinstance(bond_types, list))
    assert(isinstance(bond_pairs, list))
    del bond_ids[:]
    del bond_types[:]
    del bond_pairs[:]

    for ie in range(0, len(lines_bonds)):

        line = lines_bonds[ie].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()

        if len(line) == 0:
            continue

        tokens = ttree_lex.SplitQuotedString(line)

        if section_name == "Data Bonds AtomId AtomId":
            if len(tokens) == 2:
                bondid_n = bond_ids_offset + len(bond_ids) + 1
                bond_ids.append(prefix + str(bondid_n) + suffix)
                bond_pairs.append((ttree_lex.EscCharStrToChar(tokens[0]),
                                   ttree_lex.EscCharStrToChar(tokens[1])))
            else:
                raise(ttree_lex.InputError('Incorrect number of columns on line ' +
                                           str(ie + 1) + ' of \"' + section_name + '\" section.'))

        elif section_name == "Data Bond List":
            if len(tokens) == 3:
                bond_ids.append(ttree_lex.EscCharStrToChar(tokens[0]))
                bond_pairs.append((ttree_lex.EscCharStrToChar(tokens[1]),
                                   ttree_lex.EscCharStrToChar(tokens[2])))
            else:
                raise(ttree_lex.InputError('Incorrect number of columns on line ' +
                                           str(ie + 1) + ' of \"' + section_name + '\" section.'))

        else:
            raise(ttree_lex.InputError('Internal Error (' + g_program_name +
                                       '): Unknown section name: \"' + section_name + '\"'))

    assert(len(bond_types) == 0)
    typepattern_to_coefftypes = []

    for i in range(0, len(lines_bondsbytype)):
        line = lines_bondsbytype[i].strip()
        if '#' in line:
            icomment = line.find('#')
            line = (line[:icomment]).strip()
        if len(line) > 0:
            tokens = ttree_lex.SplitQuotedString(line)

            if (len(tokens) != 3):
                raise(ttree_lex.InputError('Error: Wrong number of columns in the \"Bonds By Type\" section of data file.\n'
                                           'Offending line:\n' +
                                           '\"' + line + '\"\n'
                                           'Expected 3 columns\n'))

            coefftype = ttree_lex.EscCharStrToChar(tokens[0])
            typepattern = []

            for typestr in tokens[1:]:
                if ttree_lex.HasRE(typestr):
                    regex_str = typestr[3:]
                    typepattern.append(re.compile(regex_str))
                else:
                    typepattern.append(ttree_lex.EscCharStrToChar(typestr))

            typepattern_to_coefftypes.append([typepattern, coefftype])

    assert(len(bond_ids) == len(bond_pairs))

    for ie in range(0, len(bond_ids)):
        bond_types.append(None)

    for ie in range(0, len(bond_ids)):
        bondid = bond_ids[ie]
        (atomid1, atomid2) = bond_pairs[ie]

        if atomid1 not in atomids2types:
            raise ttree_lex.InputError('Error: atom \"' + atomid1 + '\" not defined in \"Data Atoms\".\n'
                                       '       This usually happens when the user mistypes one of the names of the\n'
                                       '       $atoms in either a \"Data Atoms\" or \"Data Bond List\" section.\n'
                                       '       To find out where the mistake occured, search the \n'
                                       '       \"ttree_assignments.txt\" file for:\n'
                                       '       \"' + atomid1 + '\"\n')

        if atomid2 not in atomids2types:
            raise ttree_lex.InputError('Error: atom \"' + atomid2 + '\" not defined in \"Data Atoms\".\n'
                                       '       This usually happens when the user mistypes one of the names of the\n'
                                       '       $atoms in either a \"Data Atoms\" or \"Data Bond List\" section.\n'
                                       '       To find out where the mistake occured, search the \n'
                                       '       \"ttree_assignments.txt\" file for:\n'
                                       '       \"' + atomid2 + '\"\n')

        atomtype1 = atomids2types[atomid1]
        atomtype2 = atomids2types[atomid2]

        for typepattern, coefftype in typepattern_to_coefftypes:

            # use string comparisons to check if atom types match the pattern
            if (ttree_lex.MatchesAll((atomtype1, atomtype2), typepattern) or
                    ttree_lex.MatchesAll((atomtype2, atomtype1), typepattern)):
                # ("MatchesAll()" defined in "ttree_lex.py")

                bond_types[ie] = coefftype

    for ie in range(0, len(bond_ids)):
        if not bond_types[ie]:
            (atomid1, atomid2) = bond_pairs[ie]
            atomtype1 = atomids2types[atomid1]
            atomtype2 = atomids2types[atomid2]
            raise ttree_lex.InputError('Error: No bond types defined for the bond between\n'
                                       '       atoms ' + atomid1 +
                                       ' (type ' + atomtype1 + ')\n'
                                       '         and ' + atomid2 + ' (type ' + atomtype2 + ')\n'
                                       '\n'
                                       '       (If you are using a force field, then it probably means that you made a\n'
                                       '        mistake choosing at least one of these two @atom types from the list\n'
                                       '        of available atom types supplied by the force field.  To fix it, edit\n'
                                       '        the corresponding lines in the "Data Atoms" section of your LT file.)\n')


def main():
    sys.stderr.write(g_program_name + ' v' +
                     g_version_str + ' ' + g_date_str + ' ')
    if sys.version < '3':
        sys.stderr.write(' (python version < 3)\n')
    else:
        sys.stderr.write('\n')

    try:
        fname_atoms = None
        fname_bond_list = None
        fname_bondsbytype = None
        section_name = 'Data Bond List'  # (This will be replaced later.)
        atom_style = 'full'
        prefix = ''
        suffix = ''
        bond_lack_types = False

        argv = [arg for arg in sys.argv]

        # Loop over the remaining arguments not processed yet.
        # These arguments are specific to the lttree.py program
        # and are not understood by ttree.py:
        i = 1
        while i < len(argv):
            #sys.stderr.write('argv['+str(i)+'] = \"'+argv[i]+'\"\n')
            if ((argv[i].lower() == '-?') or
                    (argv[i].lower() == '--?') or
                    (argv[i].lower() == '-help') or
                    (argv[i].lower() == '-help')):
                if i + 1 >= len(argv):
                    sys.stdout.write(man_page_text + '\n')
                    sys.exit(0)

            elif argv[i].lower() == '-atoms':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                               '       text which might appear in the "Atoms" section of a LAMMPS data file.\n')
                fname_atoms = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bonds':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a file name containing lines of\n'
                                               '       text which might appear in the "Bonds" section of a LAMMPS data file.\n')
                fname_bond_list = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bond-list':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')
                    # raise ttree_lex.InputError('Error: '+argv[i]+' flag should be followed by a file name containing lines of\n'
                    #                 '       text which might appear in the "Bonds No Types" section of a LAMMPS data file.\n')
                fname_bond_list = argv[i + 1]
                section_name = "Data Bond List"
                del(argv[i:i + 2])

            elif argv[i].lower() == '-bondsbytype':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError(
                        'Error: ' + argv[i] + ' flag should be followed by a file name\n')

                    # raise ttree_lex.InputError('Error: '+argv[i]+' flag should be followed by a file name containing\n'
                    #                 '       text which might appear in the "'+section_name+' By Type" section\n'
                    #                 '       of a LAMMPS data file.\n')
                fname_bondsbytype = argv[i + 1]
                del(argv[i:i + 2])

            elif ((argv[i].lower() == '-atom-style') or
                  (argv[i].lower() == '-atom_style')):
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a an atom_style name.\n'
                                               '       (Or single quoted string which includes a space-separated\n'
                                               '       list of column names.)\n')
                atom_style = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-prefix':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a prefix string\n'
                                               '       (a string you want to appear to the left of the integer\n'
                                               '        which counts the bonded interactions you have generated.)\n')
                prefix = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i].lower() == '-suffix':
                if i + 1 >= len(argv):
                    raise ttree_lex.InputError('Error: ' + argv[i] + ' flag should be followed by a suffix string\n'
                                               '       (a string you want to appear to the right of the integer\n'
                                               '        which counts the bonded interactions you have generated.)\n')
                prefix = argv[i + 1]
                del(argv[i:i + 2])

            elif argv[i][0] == '-':
                raise ttree_lex.InputError('Error(' + g_program_name + '):\n'
                                           'Unrecogized command line argument \"' + argv[i] + '\"\n')
            else:
                i += 1

        if len(argv) != 1:
            # if there are more than 2 remaining arguments,
            problem_args = ['\"' + arg + '\"' for arg in argv[1:]]
            raise ttree_lex.InputError('Syntax Error(' + g_program_name + '):\n\n'
                                       '       Problem with argument list.\n'
                                       '       The remaining arguments are:\n\n'
                                       '         ' +
                                       (' '.join(problem_args)) + '\n\n'
                                       '       (The actual problem may be earlier in the argument list.)\n')

        bond_types = []
        bond_ids = []
        bond_pairs = []

        fatoms = open(fname_atoms, 'r')
        fbonds = open(fname_bond_list, 'r')
        fbondsbytype = open(fname_bondsbytype, 'r')
        lines_atoms = fatoms.readlines()
        lines_bonds = fbonds.readlines()
        lines_bondsbytype = fbondsbytype.readlines()
        fatoms.close()
        fbonds.close()
        fbondsbytype.close()

        LookupBondTypes(bond_types,
                        bond_ids,
                        bond_pairs,
                        lines_atoms,
                        lines_bonds,
                        lines_bondsbytype,
                        atom_style,
                        section_name,
                        prefix='',
                        suffix='')

        assert(len(bond_types) == len(bond_ids) == len(bond_pairs))

        ie = 0
        N = len(bond_types)
        for ie in range(0, N):
            sys.stdout.write(bond_ids[ie] + ' ' +
                             bond_types[ie] + ' ' +
                             bond_pairs[ie][0] + ' ' +
                             bond_pairs[ie][1] + '\n')

    except (ValueError, ttree_lex.InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Author: Andrew Jewett
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2014

import sys, io

try:
    from .dump2data import *
    from .extract_lammps_data import lammps_data_sections
except (ImportError, SystemError, ValueError):
    # not installed as a package
    from dump2data import *
    from extract_lammps_data import lammps_data_sections

g_program_name = 'raw2data.py'
g_date_str = '2016-12-21'
g_version_str = 'v0.44.0'

    #######  Main Code Below: #######
def main():
    sys.stderr.write(g_program_name + ' ' + g_version_str + ' ' + g_date_str)
    sys.stderr.write('\n')

    try:
        data_settings = DataSettings()
        misc_settings = MiscSettings()
        misc_settings.multi = False

        # First process any arguments which are specific to "raw2data.py"
        # (and remove them before passing them to dump2data.ParseArgs())
        sort_data_file_by_atom_id = False
        argv = [arg for arg in sys.argv]
        i = 1
        while i < len(argv):
            if argv[i].lower() == '-ignore-atom-id':
                sort_data_file_by_atom_id = True
                del argv[i:i+1]
            if argv[i].lower() == '-sort':
                sort_data_file_by_atom_id = False
                del argv[i:i+1]
            else:
                i += 1
            #(perhaps later I'll add some additional argumets)

        warning_strings = []
        ParseArgs(argv,
                  misc_settings,
                  data_settings,
                  warning_strings)

        frame_atom_order = []

        # The atoms in the "Atoms" section of the data file might be out 
        # of order. Extract the text from that section, and figure out the 
        # atom-ID assigned to each atom (number in the first column).
        # Then assign the these atom-ID numbers to entries in the 
        # coordinates list (frame_coords) in the same order.  We want the
        # lines of text in the coordinate file to pasted to the end of
        # the lines of text in the "Atoms" section of the data file
        # in the same order, regardless of the atom-ID numbers in that file.
        # Counterintuitively, the only way to do that is to assign the same
        # atom-ID numbers to the coordinate list in frame_coords".  So we
        # have to extract those numbers from the data file.
        in_atoms_section = False
        num_blank_lines = 0
        for line in data_settings.contents:
            ic = line.find('#')
            line = line[:ic]    #(this also removes the newline character)
            tokens = line.strip().split()
            if line.strip() == 'Atoms':
                in_atoms_section = True
            elif line.strip() in lammps_data_sections:
                in_atoms_section = False
            elif in_atoms_section:
                if len(tokens) > 0:
                    if sort_data_file_by_atom_id:
                        atomid = tokens[0]
                        frame_atom_order.append(atomid)
                        #sys.stderr.write('atomid=\"'+str(atomid)+'\"\n')
                    else:
                        frame_atom_order.append(str(len(frame_atom_order)+1))
                else:
                    num_blank_lines += 1
                    if num_blank_lines > 1:
                        in_atoms_section = False

        frame_coords = defaultdict(list)
        frame_coords_ixiyiz = defaultdict(list)
        frame_vects = defaultdict(list)
        frame_velocities = defaultdict(list)
        frame_xlo_str = frame_xhi_str = None
        frame_ylo_str = frame_yhi_str = None
        frame_zlo_str = frame_zhi_str = None
        frame_xy_str = frame_xz_str = frame_yz_str = None
        frame_natoms = -1
        frame_timestep_str = ''
        i_atomid = i_atomtype = i_molid = -1
        i_x = i_y = i_z = i_xu = i_yu = i_zu = -1
        i_xs = i_ys = i_zs = i_xsu = i_ysu = i_zsu = -1

        dump_column_names = []

        #num_frames_in = -1
        num_frames_out = 0
        finished_reading_frame = False
        read_last_frame = False

        in_coord_file = sys.stdin
        #in_coord_file = open('tmp_atom_coords.dat','r')

        read_last_frame = False
        while True:

            line = '\n'
            while (line != '') and (line.strip() == ''):
                line = in_coord_file.readline()

            if line == '':  # if EOF
                break

            #frame_vects = defaultdict(list)
            frame_coords = defaultdict(list)
            while line.strip() != '':
                n_crds = len(frame_coords)
                #sys.stdout.write("n_crds="+str(n_crds)+": \""+line.strip()+"\"\n")
                frame_coords[frame_atom_order[n_crds]] = line.split()
                #frame_vects[frame_atom_order[n_crds]] = ['0.0','0.0','0.0']
                line = in_coord_file.readline()

            # Check to see if there are any blank lines at this location in the file
            # If there are, it means we are at a new "frame" (in the trajectory).
            # Skip over any blank line(s) separating this frame from the next frame
            # so that the next time we enter the loop, we are at the beginning
            # of a new frame.
            while (line != '') and (line.strip() == ''):
                line = in_coord_file.readline()

            # Parse the DATA file specified by the user
            # and replace appropriate lines or fields with
            # the corresponding text from the input file.
            if misc_settings.multi:
                out_file_name = data_settings.file_name + '.'\
                    + str(num_frames_out)
                sys.stderr.write('  (creating file \"' + out_file_name + '\")\n')
                out_file = open(out_file_name, 'w')
            else:
                out_file = sys.stdout

            WriteFrameToData(out_file,
                             None,
                             misc_settings,
                             data_settings,
                             None,
                             frame_natoms,
                             frame_coords,
                             frame_coords_ixiyiz,
                             frame_vects,
                             frame_velocities,
                             None,
                             None,
                             frame_xlo_str, frame_xhi_str,
                             frame_ylo_str, frame_yhi_str,
                             frame_zlo_str, frame_zhi_str,
                             frame_xy_str, frame_xz_str, frame_yz_str)

    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return


if __name__ == '__main__':
    main()



# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

try:
    from .nbody_graph_search import Ugraph
except (ImportError, SystemError, ValueError):
    # not installed as a package
    from nbody_graph_search import Ugraph

# This file defines how improper interactions are generated by moltemplate.sh
# by default.  It can be overridden by supplying your own custom file
# (for example, see "opls_imp.py")

#    To find 4-body "improper" interactions,
#    (by default, most of the time), we would use this subgraph:
#           3
#           *                  1st bond connects atoms 0 and 1
#           |              =>  2nd bond connects atoms 0 and 2
#         _.*._                3rd bond connects atoms 0 and 3
#       *'  0  `*
#      1         2
#

bond_pattern = Ugraph([(0, 1), (0, 2), (0, 3)])
# (Ugraph atom indices begin at 0, not 1)


def canonical_order(match):
    """
    Before defining a new interaction, we must check to see if an
    interaction between these same 4 atoms has already been created
     (perhaps listed in a different, but equivalent order).
    If we don't check for this this, we will create many unnecessary redundant
    interactions (which can slow down he simulation).
    To avoid this, I define a "canonical_order" function which sorts the atoms
    and bonds in a way which is consistent with the symmetry of the interaction
    being generated...  Later the re-ordered list of atom and bond ids will be
    tested against the list of atom/bond ids in the matches-found-so-far,
    before it is added to the list of interactions found so far.  Note that
    the energy of an improper interactions is a function of the improper angle.
    The improper-angle is usually defined as the angle between planes formed
    by atoms 0,1,2 & 1,2,3.  (Alternately, it is sometimes defined as the
    angle between the 0,1,2 plane and atom 3.)
    This angle does not change when swapping the MIDDLE pair of atoms (1 and 2)
    (except for a change of sign, which does not matter since the energy functions
     used are typically sign invariant.  Furthermore, neither of MIDDLE pair of atoms
     are the central atom, and there are 3!=6 ways of ordering the remaining 3 atoms.)
    Consequently it does not make sense to define a separate 4-body improper
    interaction between atoms 0,1,2,3   AS WELL AS between  0,2,1,3.
    So we sort the atoms and bonds so that the second atom has a always has
    a lower atomID than the third atom.  (Later we will check to see if we
    have already defined an interaction between these 4 atoms.  If not then
    we create a new one.)

    """
    atom0 = match[0][0]
    atom1 = match[0][1]
    atom2 = match[0][2]
    atom3 = match[0][3]
    # match[1][0:2] contains the ID numbers for the 3 bonds
    bond0 = match[1][0]
    bond1 = match[1][1]
    bond2 = match[1][2]
    if atom1 <= atom2:
        # return ((atom0,atom1,atom2,atom3), (bond0,bond1,bond2))
        # But this is the same thing as:
        return match
    else:
        return ((atom0, atom2, atom1, atom3), (bond1, bond0, bond2))



try:
    from ..nbody_graph_search import Ugraph
except:
    # not installed as a module
    from nbody_graph_search import Ugraph

#    To find 4-body "improper" interactions,
#    (by default, most of the time), we would use this subgraph:
#           3
#           *                  1st bond connects atoms 0 and 1
#           |              =>  2nd bond connects atoms 0 and 2
#         _.*._                3rd bond connects atoms 0 and 3
#       *'  0  `*
#      1         2
#

bond_pattern = Ugraph([(0,1), (0,2), (0,3)])
# (Ugraph atom indices begin at 0, not 1)


def canonical_order(match):
    """
    When searching for atoms with matching bond patterns GraphMatcher
    often returns redundant results. We must define a "canonical_order"
    function which sorts the atoms and bonds in a way which is consistent
    with the type of N-body interaction being considered.
    The atoms (and bonds) in a candidate match are rearranged by the
    canonical_order().  Then the re-ordered list of atom and bond ids is
    tested against the list of atom/bond ids in the matches-found-so-far,
    before it is added to the list of interactions found so far.
    (For example, it does not make sense to define a separate 4-body improper-
    angle interaction between atoms 0, 1, 2, 3  AND 0, 2, 1, 3.
    The "improper angle" is often defined as the angle between planes formed
    by atoms 0,1,2 & 1,2,3.  Alternately, it may instead be defined as the
    angle between the 0,1,2 plane and atom 3.  Either way, this angle does
    not change when swapping the middle pair of atoms (1 and 2)
    (except for a change of sign, which does not matter since the energy functions
     used are typically sign invariant.  Furthermore, neither of OUTER pair of atoms
     are the central atom. There are 3!=6 ways of ordering the remaining 3 atoms.)
    Consequently it does not make sense to define a separate 4-body improper-
    interaction between atoms 0,1,2,3   AS WELL AS between  0,2,1,3.
    So we sort the atoms and bonds so that the first atom has a always has
    a lower atomID than the last atom.  (Later we will check to see if we
    have already defined an interaction between these 4 atoms.  If not then
    we create a new one.)

    """
    atom0 = match[0][0]
    atom1 = match[0][1]
    atom2 = match[0][2]
    atom3 = match[0][3]
    # match[1][0:2] contains the ID numbers for the 3 bonds
    bond0 = match[1][0]
    bond1 = match[1][1]
    bond2 = match[1][2]
    if atom1 <= atom2:
        #return ((atom0,atom1,atom2,atom3), (bond0, bond1, bond2))
        # But this is the same thing as:
        return match
    else:
        return ((atom0,atom2,atom1,atom3), (bond1, bond0, bond2))



try:
    from ..nbody_graph_search import Ugraph
except:
    # not installed as a module
    from nbody_graph_search import Ugraph

#    To find 4-body "improper" interactions, we would use this subgraph:
#           3
#           *                  1st bond connects atoms 0 and 1
#           |              =>  2nd bond connects atoms 0 and 2
#         _.*._                3rd bond connects atoms 0 and 3
#       *'  0  `*
#      1         2
#

bond_pattern = Ugraph([(0,1), (0,2), (0,3)])
# (Note: Ugraph atom-index counters begin at 0, not 1)


def canonical_order(match):
    """
    When searching for atoms with matching bond patterns GraphMatcher
    often returns redundant results. We must define a "canonical_order"
    function which sorts the atoms and bonds in a way which is consistent
    with the type of N-body interaction being considered.
    The atoms (and bonds) in a candidate match are rearranged by the
    canonical_order().  Then the re-ordered list of atom and bond ids is
    tested against the list of atom/bond ids in the matches-found-so-far,
    before it is added to the list of interactions found so far.  In this
    case we assume the second atom is the central atom (the "hub"), and the
    energy is invariant with respect to permutations of the other 3 atoms.
    So we arbitrarily sort these other 3 atoms in increasing order
    (as well as the bonds which connect the central atom to them).

    """
    atom0 = match[0][0]
    atom1 = match[0][1]
    atom2 = match[0][2]
    atom3 = match[0][3]
    bonds = match[1]
    ab=[(atom1,0), (atom2,1), (atom3,2)]
    ab.sort()
    return ((atom0, ab[0][0], ab[1][0], ab[2][0]),
            (bonds[ab[0][1]], bonds[ab[1][1]], bonds[ab[2][1]]))



# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013

espt_delim_atom_fields = set(["pos", "type", "v", "f", 
                              "bond", 
                              "temp", "gamma", 
                              "q", 
                              "quat", "omega", "torque", 
                              "rinertia", "fix", "unfix", "ext_force",
                              "exclude", "delete", 
                              "mass",
                              "dipm", "dip", "virtual",
                              "vs_relative", "distance", "vs_auto_relate_to"])



def LinesWSlashes(text):
    """ 
    Iterate over the lines contained in a string of text.
    Merge lines ending in backslashes.

    """

    current_line = ''
    for line in text.split('\n'):
        current_line += line
        if (len(line) > 0) and (line[-1] != '\\'):
            yield current_line
            current_line = ''
    if len(current_line) > 0:
        yield current_line




def SplitMultiDelims(line, delimiters):
    """
    Split a string into tokens using one or more (multi-character) delimiters.
    (Bug: The current version of this function does not preserve white space,
          but this should not matter.)

    """

    token = ''
    for sub_token in line.strip().split():
        if sub_token in delimiters:
            yield token
            yield sub_token
            token = ''
        elif len(token) > 0:
            token += ' ' + sub_token
        else:
            token += sub_token
    if len(token) > 0:
        yield token



def SplitAtomLine(line):
    l = []
    for token in SplitMultiDelims(line, espt_delim_atom_fields):
        l.append(token)
    return l

    # In this type of TCL command, all of the delimiters 
    #    (like 'pos', 'type', 'q', ...) 
    # are supposed to be followed by an argument.  If the last 
    # token on this line IS a delimiter, then this is a syntax error.

    if token in espt_delim_atom_fields:
        raise InputError("Error: Incomplete line:\n"
                         "\""+line+"\"\n")



def iEsptAtomCoords(tokens):
    #tokens = SplitMultiDelims(line)
    i = 0
    while i < len(tokens):
        if tokens[i] in set(['pos', 'fix', 'unfix']):
            assert(i+1 < len(tokens))
            yield i+1
            i += 1
        i += 1




def iEsptAtomVects(tokens):
    #tokens = SplitMultiDelims(line)
    i = 0
    while i < len(tokens):
        if tokens[i] in set(['dip', 'rinertia', 'v', 'f', 'omega', 'torque']):
            assert(i+1 < len(tokens))
            yield i+1
            i += 1
        i += 1


def iEsptAtomType(tokens):
    #tokens = SplitMultiDelims(line)
    i = 0
    while i < len(tokens):
        if tokens[i] == 'type':
            assert(i+1 < len(tokens))
            yield i+1
            i += 1
        i += 1

def iEsptAtomID(tokens):
    if len(tokens) > 1:
        return 1
    else:
        raise InputError("Error: Incomplete line:\n"
                         "\""+line+"\"\n")




#!/usr/bin/env python

# Author: Andrew Jewett (jewett.aij at g mail)
# License: MIT License  (See LICENSE.md)
# Copyright (c) 2013
# All rights reserved.

"""
    lttree_postprocess.py
    This is a stand-alone python script which checks the files created by
    lttree.py to insure that the standard instance-variables ($variables)
    have all been defined.  This script performs a task which is very similar
    to the task performed by lttree_check.py.  This script attempts to detect
    mistakes in the names of $atom, $bond, $angle, $dihedral, $improper, & $mol
    variables.

"""

import sys

try:
    from .lttree_styles import *
    from .ttree_lex import ExtractCatName
except (ImportError, SystemError, ValueError):
    # not installed as a package
    from lttree_styles import *
    from ttree_lex import ExtractCatName

g_program_name = __file__.split('/')[-1]  # = 'lttree_postprocess.py'
g_version_str = '0.5.1'
g_date_str = '2017-8-23'

def main():
    atom_style = 'full'
    ttree_assignments_fname = 'ttree_assignments.txt'

    defined_mols = set([])
    defined_atoms = set([])
    defined_bonds = set([])
    defined_angles = set([])
    defined_dihedrals = set([])
    defined_impropers = set([])

    g_no_check_msg = \
        '(To override this error, run moltemplate using the \"-nocheck\" argument.)\n'

    if len(sys.argv) > 1:
        for i in range(0, len(sys.argv)):
            if ((sys.argv[i].lower() == '-atomstyle') or
                    (sys.argv[i].lower() == '-atom-style') or
                    (sys.argv[i].lower() == '-atom_style')):
                if i + 1 >= len(sys.argv):
                    raise InputError('Error(' + g_program_name + '): The ' + sys.argv[i] + ' flag should be followed by a LAMMPS\n'
                                     '       atom_style name (or single quoted string containing a space-separated\n'
                                     '       list of column names such as: atom-ID atom-type q x y z molecule-ID.)\n')

                atom_style = sys.argv[i + 1]
            elif ((sys.argv[i].lower() == '-ttreeassignments') or
                  (sys.argv[i].lower() == '-ttree-assignments') or
                  (sys.argv[i].lower() == '-ttree_assignments')):
                if i + 1 >= len(sys.argv):
                    raise InputError('Error(' + g_program_name + '): The ' + sys.argv[i] + ' flag should be followed by \n'
                                     '       a file containing the variable bindings created by ttree/moltemplate.\n')
                ttree_assignments_fname = sys.argv[i + 1]
            else:
                pass  # ignore other arguments (they are intended for lttree.py)


    atom_column_names = AtomStyle2ColNames(atom_style)
    i_atomid = 0
    i_molid = -1
    for i in range(0, len(atom_column_names)):
        if atom_column_names[i].lower() == 'atom-id':
            i_atomid = i
        elif atom_column_names[i].lower() == 'molecule-id':
            i_molid = i
    i_max_column = max(i_atomid, i_molid)


    # The following variables are defined in "lttree_styles.py"
    #data_atoms="Data Atoms"
    #data_masses="Data Masses"
    #data_velocities="Data Velocities"
    #data_bonds="Data Bonds"
    #data_angles="Data Angles"
    #data_dihedrals="Data Dihedrals"
    #data_impropers="Data Impropers"


    sys.stderr.write(g_program_name + ' v' +
                     g_version_str + ' ' + g_date_str + '\n')

    try:
        # ------------ defined_atoms ------------
        try:
            f = open(data_atoms + '.template', 'r')
        except:
            raise InputError('Error(' + g_program_name + '): Unable to open file\n' +
                             '\"' + data_atoms + '.template\"\n'
                             '       for reading.  (Do your files lack a \"' +
                             data_atoms + '\" section?)\n'
                             + g_no_check_msg + '\n')

        for line_orig in f:
            ic = line_orig.find('#')
            if ic != -1:
                line = line_orig[:ic]
            else:
                line = line_orig.rstrip('\n')

            tokens = line.strip().split()
            if len(tokens) == 0:
                pass
            elif len(tokens) <= i_max_column:
                raise InputError('Error(' + g_program_name + '): The following line from\n'
                                 '     "\"' + data_atoms + '.template\" has bad format:\n\n'
                                 + line_orig + '\n'
                                 '     This my probably an internal error. (Feel free to contact the developer.)\n'
                                 + g_no_check_msg + '\n')
            else:
                defined_atoms.add(tokens[i_atomid])
                if i_molid != -1:
                    defined_mols.add(tokens[i_molid])

        f.close()

        # ------------ defined_bonds ------------
        try:
            f = open(data_bonds + '.template', 'r')

            for line_orig in f:
                ic = line_orig.find('#')
                if ic != -1:
                    line = line_orig[:ic]
                else:
                    line = line_orig.rstrip('\n')

                tokens = line.strip().split()
                if len(tokens) == 0:
                    pass
                elif len(tokens) < 4:
                    raise InputError('Error(' + g_program_name + '): The following line from\n'
                                     '     "\"' + data_bonds + '.template\" has bad format:\n\n'
                                     + line_orig + '\n'
                                     '     This my probably an internal error. (Feel free to contact the developer.)\n'
                                     + g_no_check_msg + '\n')
                else:
                    defined_bonds.add(tokens[0])
            f.close()
        except:
            pass  # Defining bonds (stored in the data_bonds file) is optional

        # ------------ defined_angles ------------
        try:
            f = open(data_angles + '.template', 'r')
            for line_orig in f:
                ic = line_orig.find('#')
                if ic != -1:
                    line = line_orig[:ic]
                else:
                    line = line_orig.rstrip('\n')

                tokens = line.strip().split()
                if len(tokens) == 0:
                    pass
                elif len(tokens) < 5:
                    raise InputError('Error(' + g_program_name + '): The following line from\n'
                                     '     "\"' + data_angles + '.template\" has bad format:\n\n'
                                     + line_orig + '\n'
                                     '     This my probably an internal error. (Feel free to contact the developer.)\n'
                                     + g_no_check_msg + '\n')
                else:
                    defined_angles.add(tokens[0])
            f.close()
        except:
            pass  # Defining angles (stored in the data_angles file) is optional

        # ------------ defined_dihedrals ------------
        try:
            f = open(data_dihedrals + '.template', 'r')
            for line_orig in f:
                ic = line_orig.find('#')
                if ic != -1:
                    line = line_orig[:ic]
                else:
                    line = line_orig.rstrip('\n')

                tokens = line.strip().split()
                if len(tokens) == 0:
                    pass
                elif len(tokens) < 6:
                    raise InputError('Error(' + g_program_name + '): The following line from\n'
                                     '     "\"' + data_dihedrals + '.template\" has bad format:\n\n'
                                     + line_orig + '\n'
                                     '     This my probably an internal error. (Feel free to contact the developer.)\n'
                                     + g_no_check_msg + '\n')
                else:
                    defined_dihedrals.add(tokens[0])
            f.close()
        except:
            # Defining dihedrals (stored in the data_dihedrals file) is optional
            pass

        # ------------ defined_impropers ------------
        try:
            f = open(data_impropers + '.template', 'r')

            for line_orig in f:
                ic = line_orig.find('#')
                if ic != -1:
                    line = line_orig[:ic]
                else:
                    line = line_orig.rstrip('\n')

                tokens = line.strip().split()
                if len(tokens) == 0:
                    pass
                elif len(tokens) < 6:
                    raise InputError('Error(' + g_program_name + '): The following line from\n'
                                     '     "\"' + data_impropers + '.template\" has bad format:\n\n'
                                     + line_orig + '\n'
                                     '     This my probably an internal error. (Feel free to contact the developer.)\n'
                                     + g_no_check_msg + '\n')
                else:
                    defined_impropers.add(tokens[0])
            f.close()
        except:
            # Defining impropers (stored in the data_impropers file) is optional
            pass

        # ---- Check ttree_assignments to make sure variables are defined ----

        try:
            f = open(ttree_assignments_fname, 'r')
        except:
            raise InputError('Error(' + g_program_name + '): Unable to open file\n' +
                             '\"' + ttree_assignments_fname + '\"\n'
                             '       for reading.  (Do your files lack a \"' +
                             data_atoms + '\" section?)\n'
                             + g_no_check_msg + '\n')

        for line_orig in f:

            ic = line_orig.find('#')
            if ic != -1:
                line = line_orig[:ic]
                usage_location_str = 'near ' + line_orig[ic + 1:]
            else:
                line = line_orig.rstrip('\n')
                usage_location_str = ''

            tokens = line.strip().split()
            if len(tokens) == 0:
                pass
            if len(tokens) > 0:
                # This file contains a list of variables of the form:
                #
                # @/atom:MoleculeType1:C    1
                # @/atom:MoleculeType1:H    2
                # @/atom:MoleculeType2:N    3
                # $/atom:molecule1:N1    1
                # $/atom:molecule1:C1    2
                #   :
                # $/atom:molecule1141:CH    13578
                # $/atom:molecule1142:N3    13579
                #   :
                # We only care about instance variables (which use the '$' prefix)
                # Lines corresponding to static variables (which use the '@' prefix)
                # are ignored during this pass.
                i_prefix = tokens[0].find('$')
                if i_prefix != -1:
                    descr_str = tokens[0][i_prefix + 1:]
                    cat_name = ExtractCatName(descr_str)

                    if ((cat_name == 'atom') and
                            (tokens[0] not in defined_atoms)):
                        raise InputError('Error(' + g_program_name + '): ' + usage_location_str + '\n' +
                                         '      Reference to undefined $atom:\n\n'
                                         '            ' + tokens[0] + '     (<--full name)\n\n' +
                                         '      (If that atom belongs to a molecule (or other subunit), make sure that\n' +
                                         '       you specified the correct path which leads to it (using / and ..))\n\n' +
                                         g_no_check_msg)

                    elif ((cat_name == 'bond') and
                          (tokens[0] not in defined_bonds)):
                        raise InputError('Error(' + g_program_name + '): ' + usage_location_str + '\n' +
                                         '      Reference to undefined $bond:\n\n'
                                         '            ' + tokens[0] + '     (<--full name)\n\n' +
                                         '      (If that bond belongs to a molecule (or other subunit), make sure that\n' +
                                         '       you specified the correct path which leads to it (using / and ..))\n\n' +
                                         g_no_check_msg)

                    elif ((cat_name == 'angle') and
                          (tokens[0] not in defined_angles)):
                        raise InputError('Error(' + g_program_name + '): ' + usage_location_str + '\n' +
                                         '      Reference to undefined $angle:\n\n' +
                                         '            ' + tokens[0] + '     (<--full name)\n\n' +
                                         '      (If that angle belongs to a molecule (or other subunit), make sure that\n' +
                                         '       you specified the correct path which leads to it (using / and ..))\n\n' +
                                         g_no_check_msg)

                    elif ((cat_name == 'dihedral') and
                          (tokens[0] not in defined_dihedrals)):
                        raise InputError('Error(' + g_program_name + '): ' + usage_location_str + '\n\n' +
                                         '      Reference to undefined $dihedral:\n\n'
                                         '            ' + tokens[0] + '     (<--full name)\n\n' +
                                         '    (If that dihedral belongs to a molecule (or other subunit), make sure that\n' +
                                         '     you specified the correct path which leads to it (using / and ..))\n\n' +
                                         g_no_check_msg)

                    elif ((cat_name == 'improper') and
                          (tokens[0] not in defined_impropers)):
                        raise InputError('Error(' + g_program_name + '): ' + usage_location_str + '\n' +
                                         '      Reference to undefined $improper:\n\n'
                                         '            ' + tokens[0] + '     (<--full name)\n\n' +
                                         '    (If that improper belongs to a molecule (or other subunit), make sure that\n' +
                                         '     you specified the correct path which leads to it (using / and ..))\n\n' +
                                         g_no_check_msg)

                    # I used to generate an error when a users defines a $mol
                    # variable but does not associate any atoms with it (or if the
                    # user systematically deletes all the atoms in that molecule),
                    # but I stopped this practice.
                    # I don't think there is any real need to complain if some
                    # molecule id numbers are undefined.  LAMMPS does not care.
                    #
                    # elif ((cat_name == 'mol') and
                    #    (tokens[0] not in defined_mols)):
                    #    raise InputError('Error('+g_program_name+'): '+usage_location_str+'\n'+
                    #                     '      Reference to undefined $mol (molecule-ID) variable:\n\n'
                    #                     '            '+tokens[0]+'     (<--full name)\n\n'+
                    #                     '    (If that molecule is part of a larger molecule, then make sure that\n'+
                    #                     '     you specified the correct path which leads to it (using / and ..))\n\n'+
                    #                     g_no_check_msg)

        f.close()

        sys.stderr.write(g_program_name + ': -- No errors detected. --\n')
        exit(0)


    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(1)

    return

if __name__ == '__main__':
    main()



try:
    from ..nbody_graph_search import Ugraph
except (SystemError, ValueError):
    # not installed as a package
    from nbody_graph_search import Ugraph


#    To find 2-body "bond" interactions, we would use this subgraph:
#
#
#       *---*           =>  one bond connects atoms 0 and 1
#       0   1
#

bond_pattern = Ugraph([(0, 1)])
# (Ugraph atom indices begin at 0, not 1)


#    The next function eliminates the redundancy between 0-1-2 and 2-1-0:
def canonical_order(match):
    """
    When searching for atoms with matching bond patterns GraphMatcher
    often returns redundant results. We must define a "canonical_order"
    function which sorts the atoms and bonds in a way which is consistent
    with the type of N-body interaction being considered.
    However, occasionally we DON'T want to modify the atom order.
    In this case, this function returns
    the original "match" argument unmodified.

    """

    return match
